import os
import platform
import json
import tempfile
import dataclasses
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Tuple, Dict, Any, Optional

from libbiomedit.lib.deserialize import deserialize

from .log import get_default_log_dir, create_logger
from ..core.crypt import open_gpg_dir, get_default_gpg_config_dir
from ..core.error import UserError
from ..core.portal_api import PortalApi
from .. import APP_NAME_SHORT


CONFIG_FILE_NAME = "config.json"
CONFIG_FILE_ENVIRON_VAR = "SETT_CONFIG_FILE"

conf_sub_dir_by_os: Dict[str, Tuple[str, ...]] = {
    "Linux": (".config",),
    "Darwin": (".config",),
    "Windows": ("AppData", "Roaming"),
}

logger = create_logger(__name__)


@dataclass(frozen=True)
class Connection:
    """dataclass holding config fo a connection (sftp / liquid files)"""

    protocol: str
    parameters: Dict[str, Any]


TMP_DIR = tempfile.gettempdir()


@dataclass
class Config:
    """dataclass holding config data"""

    dcc_portal_url: str = "https://portal.dcc.sib.swiss"
    keyserver_url: Optional[str] = "keyserver.dcc.sib.swiss"
    gpg_config_dir: str = get_default_gpg_config_dir()
    key_validation_authority_keyid: Optional[str] = "881685B5EE0FCBD3"
    sign_encrypted_data: bool = True
    always_trust_recipient_key: bool = True
    repo_url: str = "https://pypi.org"
    check_version: bool = True
    offline: bool = False
    log_dir: str = get_default_log_dir()
    error_reports: bool = True
    log_max_file_number: int = 1000
    connections: Dict[str, Connection] = dataclasses.field(default_factory=lambda: {})
    temp_dir: str = TMP_DIR
    output_dir: Optional[str] = None
    ssh_password_encoding: str = "utf_8"
    default_sender: Optional[str] = None
    gui_quit_confirmation: bool = True
    compression_level: int = 5

    def __post_init__(self):
        for url in ("dcc_portal_url", "repo_url"):
            setattr(self, url, getattr(self, url).rstrip("/"))

    @property
    def portal_api(self):
        return PortalApi(self.dcc_portal_url)

    @property
    def gpg_store(self):
        return open_gpg_dir(self.gpg_config_dir)


class ConnectionStore:
    """Connection configuration storage manager"""

    config_field_name = "connections"

    def __init__(self, config_path: Optional[str] = None):
        self.path = get_config_file() if config_path is None else config_path

    def _read(self) -> Dict:
        """Load data from config file"""
        try:
            with open(self.path) as f:
                return json.load(f)
        except FileNotFoundError:
            return {}

    def _write(self, data: Dict):
        """Write data to config file"""
        save_config(data, self.path)

    def save(self, name: str, connection: Connection, exclude: str = "pass"):
        """Save a new connection in the config file

        :param name: name of the new connection.
        :param connection: new connection object.
        :param exclude: exclude all connection fields containing `exclude` substring.
        """
        data = self._read()
        connections = data.setdefault(self.config_field_name, {})
        c = dataclasses.asdict(connection)
        pass_fields = {k for k in c["parameters"] if exclude in k}
        for f in pass_fields:
            del c["parameters"][f]
        connections[name] = c
        self._write(data)

    def delete(self, name: str):
        """Delete a new connection from the config file"""
        data = self._read()
        try:
            data.get(self.config_field_name, {}).pop(name)
            self._write(data)
        except KeyError as e:
            raise UserError(f"Connection '{name}' does not exist.") from e

    def rename(self, old: str, new: str):
        """Rename an existing connection from the config file"""
        data = self._read()
        try:
            connection = data.get(self.config_field_name, {}).pop(old)
            data[self.config_field_name][new] = connection
            self._write(data)
        except KeyError as e:
            raise UserError(f"Connection '{new}' does not exist.") from e


def load_config() -> Config:
    """Loads the config, returning a Config object"""
    cfg_dct = {}
    try:
        cfg_dct = sys_config_dict()
        cfg_dct.update(load_config_dict(get_config_file()))
    except UserError as e:
        warnings.warn(format(e))
    return deserialize(Config)(cfg_dct)


def save_config(config: Dict, path: Optional[str] = None) -> str:
    """Save config to a file
    :return: path of the created config file.
    """
    config_file = Path(get_config_file() if path is None else path)
    if not config_file.parent.is_dir():
        config_file.parent.mkdir(parents=True)

    with open(config_file, "w") as f:
        json.dump(config, f, indent=2, sort_keys=True)
    return config_file.as_posix()


def create_config() -> str:
    """Creates a new config file in the home directory's config folder
    :return: path of the created config file.
    """
    return save_config(config_to_dict(default_config()))


def config_to_dict(config: Config) -> dict:
    """Converts a Config object into a dict"""
    return dataclasses.asdict(config)


def default_config() -> Config:
    """Creates a new Config object with default values"""
    return deserialize(Config)(sys_config_dict())


def sys_config_dict() -> dict:
    """On linux only: try to load global sys config.
    If the env variable `SYSCONFIG` is set search there, else in
    `/etc/{APP_NAME_SHORT}`"""
    if platform.system() == "Linux":
        sys_cfg_dir = os.environ.get("SYSCONFIG", os.path.join("/etc", APP_NAME_SHORT))
        return load_config_dict(os.path.join(sys_cfg_dir, CONFIG_FILE_NAME))
    return {}


def load_config_dict(path: str) -> dict:
    """Load raw config as a dict"""
    try:
        with open(path) as f:
            return json.load(f)
    except FileNotFoundError:
        return {}
    except json.decoder.JSONDecodeError as e:
        raise UserError(f"Failed to load configuration from '{path}'. {e}") from e


def get_config_file() -> str:
    """Retrieve the platform-specific path of the config file. If the user has
    the correct config file path environmental variable defined in their
    current environment, this file gets used instead.

    :return: path of the config.
    :raises UserError:
    """
    # Case 1: a config file path environmental variable is defined.
    if CONFIG_FILE_ENVIRON_VAR in os.environ:
        config_file = os.environ[CONFIG_FILE_ENVIRON_VAR]
        if os.path.isdir(config_file):
            raise UserError(
                f"Environmental variable {CONFIG_FILE_ENVIRON_VAR} "
                f"must point to a file, not a directory "
                f"[{config_file}]."
            )
        return config_file

    # Case 2: use the default platform-specific config file.
    return os.path.join(get_config_dir(), CONFIG_FILE_NAME)


def get_config_dir() -> str:
    """Return platform specific default config direcory."""
    return os.path.join(
        os.path.expanduser("~"), *conf_sub_dir_by_os[platform.system()], APP_NAME_SHORT
    )



from typing import Sequence

from .upload_keys import upload_keys
from ..core.crypt import verify_key_length, search_pub_key
from ..core.error import UserError
from ..utils.log import create_logger
from ..utils.config import Config


logger = create_logger(__name__)


def request_sigs(key_ids: Sequence[str], *, config=Config):
    """Requests signatures"""
    if config.offline:
        raise UserError("Requesting key signature is not possible in offline mode.")
    keys = frozenset(search_pub_key(k, config.gpg_store, sigs=False) for k in key_ids)
    for key in keys:
        verify_key_length(key)
    for key in keys:
        upload_keys((key.fingerprint,), config=config)
        logger.info("Sending a request for '%s'", key.key_id)
        config.portal_api.request_key_signature(key.key_id)



from typing import Sequence, Optional
from abc import ABC, abstractmethod
from ..utils.progress import ProgressInterface


class Protocol(ABC):
    @abstractmethod
    def upload(
        self,
        files: Sequence[str],
        progress: Optional[ProgressInterface] = None,
    ):
        ...

    def required_password_args(self) -> Sequence[str]:  # pylint: disable=no-self-use
        return ()



class UserError(Exception):
    """Base error class for errors which can be handled by the end user

    This is for error messages addressed to the end user.
    Errors which should be handled internally shall not be raised
    as this type of error.
    Use it only when you are sure it is has a end user readable message.
    All other errors should cause an ordinary traceback such that the
    cause can be debugged by the developers"""



from typing import Optional
from urllib.error import HTTPError, URLError
import base64
import json
from functools import wraps

from .request import post, urlopen
from .metadata import MetaData
from .error import UserError


class PortalApi:
    sign_request = "/backend/pgpkey/sign-request/"
    dpkg_check = "/backend/data-package/check/"

    def __init__(self, base_url):
        self.base_url = base_url

    @staticmethod
    def auth_header(username: Optional[str] = None, password: Optional[str] = None):
        auth_str = (
            username
            and password
            and base64.b64encode(":".join((username, password)).encode()).decode()
        )
        return {"Authorization": f"Basic {auth_str}"} if auth_str else {}

    def verify_transfer(self, metadata: MetaData, filename: str) -> str:
        """Verify transfer_id using external API.

        Return project_id for a valid transfer_id.
        """
        data = json.dumps(
            {"file_name": filename, "metadata": json.dumps(MetaData.asdict(metadata))}
        ).encode()
        response = handleErrors(post)(self.base_url + self.dpkg_check, data)
        try:
            return json.loads(response)["project_id"]
        except json.decoder.JSONDecodeError as e:
            raise UserError(
                "PortalApi.verify_transfer: got invalid json from portal: "
                + response.decode()
            ) from e
        except KeyError as e:
            raise UserError(
                "PortalApi.verify_transfer: json response from portal does "
                "not include field `project_id`: " + response.decode()
            ) from e

    def request_key_signature(self, keyid: str) -> bool:
        """Request key signature"""
        url = self.base_url + self.sign_request
        if not url.lower().startswith("http"):
            raise ValueError(f"Invalid scheme: '{url}'")
        data = f"pgpkey_id={keyid}".encode()
        return handleErrors(get)(url, data)


def get(url, data):
    with urlopen(url, data):
        return True


def handleErrors(f):
    @wraps(f)
    def _f(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except HTTPError as e:
            msg_raw = e.read()
            try:
                msg = json.loads(msg_raw)["detail"]
            except (json.decoder.JSONDecodeError, KeyError, TypeError):
                msg = msg_raw.decode()
            raise UserError("PortalApi: " + msg) from e
        except URLError:
            raise UserError("Could not connect to the server.") from None

    return _f



import io
import json
import os
from functools import partial
from pathlib import Path
from typing import List, Optional, Callable, Tuple, Union

from ..utils.config import Config
from ..utils.log import create_logger, log_timing
from ..core.archive import (
    write_tar,
    ArchiveInMemoryFile,
    ArchiveFile,
    ArchiveFileBase,
    METADATA_FILE,
    METADATA_FILE_SIG,
    DATA_FILE_ENCRYPTED,
    CHECKSUM_FILE,
    CONTENT_FOLDER,
)
from ..core.filesystem import (
    search_files_recursively,
    check_file_read_permission,
    delete_files,
)
from ..core.crypt import (
    retrieve_keys,
    encrypt_and_sign,
    detach_sign_file,
    check_password,
)
from ..core.checksum import write_checksums, compute_checksum_on_write
from ..core.error import UserError
from ..core.metadata import MetaData, alnum_str, Purpose, HexStr1024, HexStr256
from ..core.secret import Secret
from ..utils.progress import ProgressInterface, subprogress, progress_file_iter

DATE_FMT_FILENAME = "%Y%m%dT%H%M%S"
logger = create_logger(__name__)
output_name_str = alnum_str(min_length=1, max_length=50, allow_dots=True)


def path_str(directory: bool = False, writable: bool = False) -> Callable:
    """Generate a 'type definition' function that will check that a string is
    a valid path (file or directory).
    :param directory: if True, the path must be a directory.
    :param writable: if True, the user must have write access to the path.
    :returns: type check function.
    :raises ValueError:
    """

    def _path_str(path_to_check):
        path = Path(path_to_check)
        if not path.exists():
            raise ValueError(
                f"Invalid path: '{path_to_check}'. " f"Path does not exist."
            )
        if directory and not path.is_dir():
            raise ValueError(
                f"Invalid path: '{path_to_check}'. " f"Path is not a directory."
            )
        if writable and not os.access(path.as_posix(), os.W_OK):
            raise ValueError(
                f"Invalid path: '{path_to_check}'. " f"Path is not writable."
            )
        return path_to_check

    return _path_str


output_path_str = path_str(directory=True, writable=True)


def check_arg_value(
    arg_value: Union[str, int], arg_name: str, arg_type: Callable
) -> None:
    """Verify that the value of variable arg_value that is named arg_name is
    following type arg_type. arg_type is a function that does a check
    of the type of the variable.
    """
    try:
        arg_type(arg_value)
    except ValueError as e:
        raise UserError(f"Invalid value for argument '{arg_name}': {e}") from e


def check_integer_in_range(min_value: int, max_value: int) -> Callable:
    """Generate a function that will check that its input value is an integer,
    and that this integer is in the range [min_value:max_value].
    """

    def _integer_in_range(value_to_check: Union[int, str]) -> None:
        try:
            value_to_check = int(value_to_check)
        except ValueError as e:
            raise ValueError("value must be an integer") from e

        if value_to_check < min_value or value_to_check > max_value:
            raise ValueError(
                f"value must be in the range " f"[{min_value}-{max_value}]"
            )

    return _integer_in_range


check_compression_level = check_integer_in_range(min_value=0, max_value=9)


# pylint: disable=too-many-statements
@log_timing(logger)
def encrypt(
    files: List[str],
    *,
    config: Config,
    recipient: List[str],
    dtr_id: Optional[int] = None,
    sender: Optional[str] = None,
    passphrase: Optional[Secret] = None,
    output_name: Optional[str] = None,
    dry_run: bool = False,
    offline: bool = False,
    verify_dtr: bool = True,
    compression_level: Optional[int] = None,
    purpose: Optional[Purpose] = None,
    progress: ProgressInterface = None,
) -> Optional[str]:
    """Compress and encrypt files and/or directories.

    Returns the file name of the created package
    """

    # Retrive non-specified, optional argument values from config.
    if sender is None:
        sender = config.default_sender or config.gpg_store.default_key()
        if sender is None:
            raise UserError("Sender not specified with no default sender.")
    if compression_level is None:
        compression_level = config.compression_level
    check_arg_value(
        arg_value=compression_level,
        arg_name="compression level",
        arg_type=check_compression_level,
    )

    if not files:
        raise UserError("Empty file list")

    files_to_encrypt = list(search_files_recursively(files))

    with logger.log_task("Input data check"):
        check_file_read_permission(files_to_encrypt)
        # Retrieve the lowest common directory of all input files/directories.
        root_dir = os.path.commonpath(
            [Path(x).absolute().parent.as_posix() for x in files]
        )
        if dtr_id is None and verify_dtr:
            raise UserError("DTR (Data Transfer Request) ID is missing.")
        if output_name:
            check_arg_value(Path(output_name).name, "output name", output_name_str)
            check_arg_value(
                Path(output_name).parent.as_posix(),
                "path in output name",
                output_path_str,
            )

    with logger.log_task("Retrieve sender and recipient GnuPG keys"):
        # Retrieve the sender's public and private keys, as well as the
        # recipient's public key. Here is what these keys are needed for:
        #  - sender private key: needed to sign the encrypted data.
        #  - sender public key : will be checked to make sure it is signed by
        #       the DCC. Only private keys with a matching public key that is
        #       signed by the DCC will be allowed to be used.
        #  - recipient public key: needed to encrypt the data.
        #
        # The sender/recipient information can be either an email, a keyID or
        # a full fingerprint.
        sender_pub_key, recipients_pub_key = retrieve_keys(
            gpg_store=config.gpg_store,
            sender=sender,
            recipients=recipient,
            validation_keyid=config.key_validation_authority_keyid,
            keyserver_url=config.keyserver_url,
            with_refresh_keys=not offline,
        )
        logger.info(
            "found the following keys:\n%s\n%s",
            f"sender key: {sender_pub_key.uids[0]}, "
            f"key ID: {sender_pub_key.key_id}",
            "\n".join(
                [
                    f"recipient key: {key.uids[0]}, " f"key ID: {key.key_id}"
                    for key in recipients_pub_key
                ]
            ),
        )

        if dtr_id is not None:
            if purpose is None:
                raise UserError("DTR ID specified but `purpose` is missing")
            if not offline and verify_dtr:
                try:
                    project_id = config.portal_api.verify_transfer(
                        metadata=MetaData(
                            transfer_id=dtr_id,
                            sender=HexStr1024(sender_pub_key.fingerprint),
                            recipients=[
                                HexStr1024(key.fingerprint)
                                for key in recipients_pub_key
                            ],
                            checksum=HexStr256("0" * 64),
                            purpose=purpose,
                        ),
                        filename="missing",
                    )
                    logger.info(
                        "DTR ID '%s' is valid for project '%s'", dtr_id, project_id
                    )
                except RuntimeError as e:
                    raise UserError(format(e)) from e
    if dry_run:
        logger.info("dry run completed successfully.")
        return None

    # Check if gpg key password is correct
    if config.sign_encrypted_data:
        check_password(
            sender_pub_key.fingerprint, enforce_passphrase(passphrase), config.gpg_store
        )

    archive_paths = [
        os.path.join(CONTENT_FOLDER, os.path.relpath(f, start=root_dir))
        for f in files_to_encrypt
    ]

    with logger.log_task("Compute sha256 checksum on input files"):
        # Write input file checksums to a file that will be added to the
        # encrypted .tar.gz archive. This information must be encrypted as
        # file names sometimes contain information about their content.
        with subprogress(progress, step_completion_increase=0.2) as scaled_progress:
            checksums = write_checksums(
                zip(
                    archive_paths,
                    progress_file_iter(
                        files=files_to_encrypt, mode="rb", progress=scaled_progress
                    ),
                )
            )

    with logger.log_task("Compress and encrypt input data " "[this can take a while]"):
        # Encryption is done with the recipient's public key and the optional
        # signing with the user's (i.e sender) private key. The user's private
        # key passphrase is needed to sign the encrypted file.
        encrypted_file = os.path.join(config.temp_dir, DATA_FILE_ENCRYPTED)
        encrypted_checksum_buf = io.StringIO()
        with subprogress(progress, step_completion_increase=0.75) as scaled_progress:
            # Create a tar archive containing all input files
            archive_content: Tuple[ArchiveFileBase, ...] = (
                ArchiveInMemoryFile(CHECKSUM_FILE, checksums),
            ) + tuple(
                ArchiveFile(a_path, f)
                for a_path, f in zip(
                    archive_paths,
                    progress_file_iter(
                        files=files_to_encrypt, mode="rb", progress=scaled_progress
                    ),
                )
            )
            with open(encrypted_file, "wb") as fout:
                encrypt_and_sign(
                    source=partial(
                        write_tar,
                        archive_content,
                        compress_level=compression_level,
                        compress_algo="gz",
                    ),
                    output=partial(
                        compute_checksum_on_write,
                        fout=fout,
                        checksum_buffer=encrypted_checksum_buf,
                    ),
                    gpg_store=config.gpg_store,
                    recipients_fingerprint=[
                        key.fingerprint for key in recipients_pub_key
                    ],
                    signature_fingerprint=sender_pub_key.fingerprint
                    if config.sign_encrypted_data
                    else None,
                    passphrase=enforce_passphrase(passphrase)
                    if config.sign_encrypted_data
                    else None,
                    always_trust=config.always_trust_recipient_key,
                )
        encrypted_checksum = encrypted_checksum_buf.read()

    try:
        with logger.log_task("Generate metadata file"):
            # Create a dictionary with all the info we want to store in the
            # .json file, then pass this dictionary to json.dump that will
            # convert it to a json file.
            # Use indent=4 to make the output file easier on the eye.
            metadata = MetaData(
                transfer_id=dtr_id,
                sender=HexStr1024(sender_pub_key.fingerprint),
                recipients=[HexStr1024(key.fingerprint) for key in recipients_pub_key],
                purpose=purpose,
                checksum=HexStr256(encrypted_checksum),
                compression_algorithm="gzip" if compression_level > 0 else "",
            )
            metadata_bytes = json.dumps(MetaData.asdict(metadata), indent=4).encode()
            metadata_signature_bytes = b""
            if config.sign_encrypted_data:
                metadata_signature_bytes = detach_sign_file(
                    metadata_bytes,
                    sender_pub_key.fingerprint,
                    enforce_passphrase(passphrase),
                    config.gpg_store,
                )

        with logger.log_task("Generate final tarball"):
            # The default value for the output name is based on date and time
            # when the script is being run.
            # Example output name is "20191011T145012".
            output_name = form_output_tar_name(
                output_name, default=metadata.timestamp.strftime(DATE_FMT_FILENAME)
            )
            in_memory_files = (
                (METADATA_FILE, metadata_bytes),
                (METADATA_FILE_SIG, metadata_signature_bytes),
            )
            with subprogress(
                progress, step_completion_increase=0.05
            ) as scaled_progress:
                archive_content = tuple(
                    ArchiveInMemoryFile(path, content)
                    for path, content in in_memory_files
                ) + tuple(
                    ArchiveFile(os.path.relpath(f.name, start=config.temp_dir), f)
                    for f in progress_file_iter(
                        files=[encrypted_file], mode="rb", progress=scaled_progress
                    )
                )
                write_tar(content=archive_content, output=output_name, compress_level=0)
    finally:
        delete_files(encrypted_file)

    logger.info("Completed data encryption: %s", output_name)
    return output_name


def enforce_passphrase(passphrase: Optional[Secret]) -> Secret:
    if passphrase is None:
        raise ValueError("No password given")
    return passphrase


def form_output_tar_name(output_name: Optional[str], default: str) -> str:
    """Define the path + name of the output tarball file of the encrypt
    workflow. If output_name does not contain any path information, the
    output directory is set to the current working directory.
    :param output_name: name or path + name of output tarball.
    :param default: default file name to use if output_name is None or output_name is a folder
    :return: path and name of the output tarball file.
    :raises UserError:
    """
    if output_name is None:
        output_name = default
    if Path(output_name).is_dir():
        output_name = os.path.join(output_name, default)
    # Add '.tar' extension to output name if needed.
    if not output_name.endswith(".tar"):
        output_name = output_name + ".tar"

    # If output_name does not contain any path info, the output path is
    # set to the current working directory.
    basename = Path(output_name).name
    if basename == output_name:
        output_dir = Path.cwd()
    else:
        output_dir = Path(output_name).parent

    # Verify that user has write permission to output directory. We do this
    # check now because the procedure of compression + encryption can take
    # a long time and we want to be able to warn the user immediately.
    if not output_dir.is_dir():
        raise UserError(f"output directory does not exist: {output_dir}")
    if not os.access(output_dir.as_posix(), os.W_OK):
        raise UserError(f"no write permission on directory: {output_dir}")

    return output_dir.joinpath(basename).as_posix()



from typing import Optional, Dict

from .component import PathInput, FieldMapping, create_slider, LineEdit, SpinBox
from .pyside import QtCore, QtGui, QtWidgets
from ..utils import config
from ..utils.config import Config
from ..utils.validation import REGEX_URL


LABEL_COMPRESSION_LEVEL = "Compression level"
STATUS_TIP_COMPRESSION_LEVEL = (
    "Compression level used in data encryption, from 0 (no compression) to "
    "9 (highest). Higher compression levels require more computing time."
)


def config_to_settings(cfg: Config) -> Dict:
    config_dict = config.config_to_dict(cfg)
    del config_dict["connections"]  # connections are handled in "Transfer"
    return config_dict


class SettingsTab(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.app_data = self.parent().app_data
        self.config = self.app_data.config

        self.field_mapping = FieldMapping()
        self.layout: QtWidgets.QGridLayout = QtWidgets.QGridLayout()

        self.persist_btn_text = "Save to config file"
        self._add_header()
        self._add_settings()
        self._add_footer()
        self._load_config()

        self.setLayout(self.layout)

    def _load_config(self):
        """Overwrites contents of widgets with values in current config"""
        settings_dict = config_to_settings(self.config)
        self.field_mapping.load(settings_dict)

    def _update_config(self, cfg: Config):
        """Replaces the `config` with `cfg` and updates the contents of the widgets in the UI"""
        self.app_data.config = cfg
        self.config = cfg
        self._load_config()

    def _add_header(self):
        top_label = QtWidgets.QLabel(
            "Changes you make are applied instantly during the current session.\n"
            f"To persist changes across restarts, make sure to click the '{self.persist_btn_text}' "
            f"button at the bottom."
        )
        top_label.setWordWrap(True)
        self.layout.addWidget(
            top_label,
            0,
            0,
            1,
            2,
        )

        reset_btn = QtWidgets.QPushButton("Reset", self)
        reset_btn.setStatusTip("Resets to your last persisted settings")
        reset_btn.clicked.connect(lambda: self._update_config(config.load_config()))
        self.layout.addWidget(reset_btn, 0, 2)

        defaults_btn = QtWidgets.QPushButton("Restore defaults", self)
        defaults_btn.setStatusTip("Reset all settings to their default values")
        defaults_btn.clicked.connect(
            lambda: self._update_config(config.default_config())
        )
        self.layout.addWidget(defaults_btn, 0, 3)

    def _add_settings(self):
        encrypt_decrypt_group = self._add_group("Data encryption / decryption settings")

        self._add_widget_int_range(
            encrypt_decrypt_group,
            LABEL_COMPRESSION_LEVEL,
            "compression_level",
            STATUS_TIP_COMPRESSION_LEVEL,
            0,
            9,
        )

        self._add_widget_str(
            encrypt_decrypt_group,
            "Default sender",
            "default_sender",
            "Default sender fingerprint for encryption",
        )

        self._add_widget_bool(
            encrypt_decrypt_group,
            "Sign encrypted data",
            "sign_encrypted_data",
            "Whether encrypted data should be signed with sender's key",
        )

        self._add_widget_bool(
            encrypt_decrypt_group,
            "Always trust recipient key",
            "always_trust_recipient_key",
            "If unchecked, the encryption key must be signed by the local user",
        )

        self._add_widget_path(
            encrypt_decrypt_group,
            "Output directory",
            "output_dir",
            "Default output directory, relevant for encryption/decryption",
            True,
        )

        self._add_widget_bool(
            encrypt_decrypt_group,
            "Offline mode",
            "offline",
            "In offline mode, sett will not make any network connections: "
            "DTR verification and automatic PGP key downloading/updating is disabled.",
        )

        pgp_keys_group = self._add_group("PGP keys")

        self._add_widget_str(
            pgp_keys_group,
            "Key validation authority key ID",
            "key_validation_authority_keyid",
            "ID of the authority key. If a value is specified, "
            "then only keys signed by this authority could be used.",
        )

        self._add_widget_path(
            pgp_keys_group,
            "GPG config directory",
            "gpg_config_dir",
            "Path of the directory where GnuPG stores its configuration",
            True,
        )

        # no validation, as specifying http:// is optional
        self._add_widget_str(
            pgp_keys_group,
            "Keyserver URL",
            "keyserver_url",
            "URL of the keyserver: used for publishing/fetching public PGP keys.",
        )

        data_transfer_group = self._add_group("Data transfer")

        self._add_widget_str(
            data_transfer_group,
            "SSH password encoding",
            "ssh_password_encoding",
            "Character encoding used for the SSH key password",
        )

        self._add_widget_str(
            data_transfer_group,
            "DCC portal URL",
            "dcc_portal_url",
            "URL of portal instance. "
            "The portal is used for key signing and DTR (Data Transfer Request) validation.",
            REGEX_URL,
        )

        sett_updates_group = self._add_group("sett updates")

        self._add_widget_str(
            sett_updates_group,
            "Repo URL",
            "repo_url",
            "Python package repository, used when looking for updates",
            REGEX_URL,
        )

        self._add_widget_bool(
            sett_updates_group,
            "Check version",
            "check_version",
            "Check whether you have the latest version of sett on startup",
        )

        misc_log_group = self._add_group("Miscellaneous and logs")

        self._add_widget_bool(
            misc_log_group,
            "Quit confirmation",
            "gui_quit_confirmation",
            "Ask for confirmation before closing the application",
        )

        self._add_widget_path(
            misc_log_group,
            "Temp directory",
            "temp_dir",
            "Path to sett working directory",
            True,
        )

        self._add_widget_path(
            misc_log_group,
            "Log directory",
            "log_dir",
            "Path to log files directory",
            True,
        )

        self._add_widget_int(
            misc_log_group,
            "Log max. file number",
            "log_max_file_number",
            "Maximum number of logfiles to keep as backup",
            1,
        )

        self._add_widget_bool(
            misc_log_group,
            "Create error reports",
            "error_reports",
            "Write an error report if some error happens",
        )

    def _add_footer(self):
        persist_btn = QtWidgets.QPushButton(self.persist_btn_text, self)
        persist_btn.setIcon(QtGui.QIcon(":icon/feather/save.png"))
        persist_btn.setStatusTip("Saves your current settings to the config file")
        persist_btn.clicked.connect(
            lambda: config.save_config(config.config_to_dict(self.config))
        )
        self.layout.addWidget(persist_btn, self.layout.rowCount() + 1, 0, 1, 4)

    def _set_config(self, key: str, value):
        return setattr(self.config, key, value)

    def _get_config(self, key: str):
        return getattr(self.config, key)

    def _add_group(self, name: str) -> QtWidgets.QGridLayout:
        group_box = QtWidgets.QGroupBox(name)
        self.layout.addWidget(group_box, self.layout.rowCount() + 1, 0, 1, 4)
        layout = QtWidgets.QGridLayout(group_box)
        layout.setColumnMinimumWidth(0, 180)
        group_box.setLayout(layout)
        return layout

    def _add_widget_str(
        self,
        layout: QtWidgets.QGridLayout,
        label: str,
        config_key: str,
        status_tip: str,
        regex: Optional[str] = None,
    ):
        widget = self.field_mapping.bind_text(config_key, LineEdit())
        widget.setStatusTip(status_tip)
        if regex:
            widget.setValidator(
                QtGui.QRegularExpressionValidator(QtCore.QRegularExpression(regex))
            )

        widget.textChanged.connect(lambda text: self._set_config(config_key, text))

        row = layout.rowCount() + 1
        layout.addWidget(QtWidgets.QLabel(label), row, 0)
        layout.addWidget(
            widget,
            row,
            1,
        )

    def _add_widget_path(
        self,
        layout: QtWidgets.QGridLayout,
        label: str,
        config_key: str,
        status_tip: str,
        directory: bool,
    ):
        widget = self.field_mapping.bind_path(
            config_key, PathInput(directory=directory, path=None)
        )
        widget.status_tip = status_tip

        widget.on_path_change(
            lambda path: self._set_config(config_key, "" if path is None else str(path))
        )

        row = layout.rowCount() + 1
        layout.addWidget(QtWidgets.QLabel(label), row, 0)
        layout.addWidget(widget.text, row, 1)
        layout.addWidget(widget.btn, row, 2)
        layout.addWidget(widget.btn_clear, row, 3)

    def _add_widget_bool(
        self,
        layout: QtWidgets.QGridLayout,
        label: str,
        config_key: str,
        status_tip: str,
    ):
        widget = self.field_mapping.bind_checked(config_key, QtWidgets.QCheckBox())
        widget.setStatusTip(status_tip)

        widget.stateChanged.connect(
            lambda state: self._set_config(config_key, state == QtCore.Qt.Checked)
        )

        row = layout.rowCount() + 1
        layout.addWidget(QtWidgets.QLabel(label), row, 0)
        layout.addWidget(
            widget,
            row,
            1,
        )

    def _add_widget_int(
        self,
        layout: QtWidgets.QGridLayout,
        label: str,
        config_key: str,
        status_tip: str,
        minimum: Optional[int] = None,
        maximum: Optional[int] = None,
    ):
        widget = self.field_mapping.bind_value(config_key, SpinBox())
        widget.setStatusTip(status_tip)
        if minimum is not None:
            widget.setMinimum(minimum)
        if maximum is not None:
            widget.setMaximum(maximum)

        widget.valueChanged.connect(lambda value: self._set_config(config_key, value))

        row = layout.rowCount() + 1
        layout.addWidget(QtWidgets.QLabel(label), row, 0)
        layout.addWidget(
            widget,
            row,
            1,
        )

    def _add_widget_int_range(
        self,
        layout: QtWidgets.QGridLayout,
        label: str,
        config_key: str,
        status_tip: str,
        minimum: int,
        maximum: int,
    ):
        def on_change(value: int):
            self._set_config(config_key, value)

        slider, slider_value = create_slider(
            minimum=minimum,
            maximum=maximum,
            initial_value=self._get_config(config_key),
            status_tip=status_tip,
            on_change=on_change,
            show_ticks=True,
        )
        self.field_mapping.bind_value(config_key, slider)
        row = layout.rowCount() + 1
        layout.addWidget(QtWidgets.QLabel(label), row, 0)
        layout.addWidget(slider, row, 1)
        layout.addWidget(slider_value, row, 2)



import io
import platform
from functools import wraps
import logging
import logging.handlers
from typing import Callable, Tuple, Dict, cast, ContextManager, Optional
from pathlib import Path
import time
from contextlib import contextmanager

from .. import RUNTIME_INFO

LOG_FORMAT = "[%(levelname)s] %(message)s"
LOG_FORMAT_WITH_TIME = "[%(asctime)s] " + LOG_FORMAT


@contextmanager
def log_task(
    message: str,
    ok_logger: Callable = logging.info,
    err_logger: Callable = logging.error,
):
    """Context manager / decorator that logs a user-defined message with level INFO upon
    entering the context manager, and either "OK" or "FAILED" upon exiting
    depending on whether exceptions occurred during its body or not.

    : param message: message to log with INFO level when entering the
        context manager.
    : param ok_logger: method associated with the logger level at which the
        message string should be logged. E.g. logging.info will log the message
        at the INFO level.
    : param err_logger: same as above but for logger level associated with
        errors.
    """
    ok_logger(message)
    try:
        yield None
    except Exception:
        err_logger("FAILED")
        raise
    ok_logger("OK")


def log_timing(logger):
    """A function decorator, logging processing time taken by the contained code"""

    def decorator(f):
        @wraps(f)
        def wrapped_f(*args, **kwargs):
            tic = time.perf_counter()
            r = f(*args, **kwargs)
            tac = time.perf_counter()
            logger.info(f"{f.__name__!r} took {sec_to_hours(int(tac - tic))}")
            return r

        return wrapped_f

    return decorator


def sec_to_hours(seconds: int) -> str:
    """Converts given seconds into something like '9h 36mn 38s'"""
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    start = 0 if h else 1 if m else 2
    return " ".join(
        f"{n}{u}" for n, u in zip((h, m, s)[start:], ("h", "m", "s")[start:])
    )


def log_runtime_info(logger: logging.Logger):
    """Logs:
    - application version
    - versions of the main dependencies
    - version of the operating system where the application is running
    """
    logger.info(RUNTIME_INFO)


class LoggerExt(logging.Logger):
    """Logger class with additional functionality"""

    def log_task(self, message: str) -> ContextManager:
        """A context manager to log a task which can potentially fail"""
        return log_task(message, ok_logger=self.info, err_logger=self.error)


logging.setLoggerClass(LoggerExt)


class StringIoHandler(logging.Handler):
    """Logging handler logging to a StringIo."""

    def __init__(self, *args, **kwargs):
        self.log = io.StringIO()
        super().__init__(*args, **kwargs)

    def emit(self, record):
        self.log.write(self.formatter.format(record))
        self.log.write("\n")


@contextmanager
def log_to_memory(logger: Optional[logging.Logger] = None):
    mem_handler = StringIoHandler()
    mem_handler.setFormatter(logging.Formatter(LOG_FORMAT_WITH_TIME))
    if logger is None:
        logger = logging.getLogger()
    logger.addHandler(mem_handler)
    saved_level = logger.level
    logger.setLevel(logging.NOTSET)
    yield mem_handler.log
    logger.removeHandler(mem_handler)
    logger.setLevel(saved_level)


def create_logger(name, logger_level: int = logging.DEBUG) -> LoggerExt:
    """Instantiate and return a new logger object with the user selected name
     and logging level (e.g. DEBUG, INFO).

    :param name: name of logger.
    :param logger_level: level of logger. Any messages under the selected
        level will not be printed to screen and logfile.
    :return: logger object
    """
    logger = logging.getLogger(name)
    logger.setLevel(logger_level)
    return cast(LoggerExt, logger)  # We called logging.setLoggerClass


def log_to_rotating_file(
    log_dir: str,
    logger: Optional[logging.Logger] = None,
    logger_level: int = logging.DEBUG,
    file_max_number: int = 1000,
    file_max_size: int = 1000000,
) -> None:
    """Add a rotating file handler to the specified logger. When the logfile
    reaches file_max_size, it gets renamed (a number is appended) and a new
    logfile is started. When file_max_number is exceeded, the oldest backup
    file is deleted.
    Logging to a file can be disabled, by setting `file_max_number = 0` or
    `file_max_size = 0`.

    :param logger: logger object to which the file handler should be added.
    :param log_dir: directory where to write the logfile.
    :param logger_level: level of logger (e.g. logging.DEBUG, logging.INFO).
        Messages under the selected level will be ignored by the handler.
    :param file_max_number: maximum number of logfiles to keep as backup.
    :param file_max_size: maximum size in bytes of an individual logfile.
    """
    if file_max_number == 0 or file_max_size == 0:
        return
    if logger is None:
        logger = logging.getLogger()
    # Create log directory if it does not exist yet.
    log_dir_path = Path(log_dir)
    if not log_dir_path.is_dir():
        try:
            log_dir_path.mkdir(parents=True)
        except PermissionError:
            logger.error("unable to create directory %s.", log_dir_path.as_posix())

    # Setup handler for logging info to file. The file logging format adds
    # the data and time at the start of each logged message.
    log_file = log_dir_path.joinpath("log")
    file_handler = logging.handlers.RotatingFileHandler(
        filename=log_file.as_posix(),
        mode="a",
        maxBytes=file_max_size,
        backupCount=file_max_number - 1,
    )
    file_handler.setFormatter(logging.Formatter(LOG_FORMAT_WITH_TIME))
    file_handler.setLevel(logger_level)
    logger.addHandler(file_handler)


def log_to_stream(
    logger: Optional[logging.Logger] = None, logger_level: int = logging.INFO
) -> None:
    """Add a stream handler to display logging info in console.

    :param logger: logger object to which the stream handler should be added.
    :param logger_level: level of logger (e.g. logging.DEBUG, logging.INFO).
        Messages under the selected level will be ignored by the handler.
    """
    if logger is None:
        logger = logging.getLogger()
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter(LOG_FORMAT))
    console_handler.setLevel(logger_level)
    logger.addHandler(console_handler)


def get_default_log_dir() -> str:
    """Returns a string with the default location of the directory where log
    files are stored on the host operating system.
    """
    default_log_dir_by_os: Dict[str, Tuple[str, ...]] = {
        "Linux": (".local/var/log/sett",),
        "Darwin": (".local/var/log/sett",),
        "Windows": ("AppData", "Roaming", "sett"),
    }
    return Path.home().joinpath(*default_log_dir_by_os[platform.system()]).as_posix()



import html
from pathlib import Path
from typing import Iterator, Optional, Sequence, Callable

from sett.core.error import UserError
from .component import SelectionAction, ToolBar, LineEdit, Action, warning_callback
from .parallel import Worker
from .pyside import QtCore, QtGui, QtWidgets, QAction
from ..core import crypt
from ..utils.config import Config
from ..workflows import (
    upload_keys as upload_keys_workflow,
    request_sigs as request_sigs_workflow,
)

Key = crypt.gpg.Key


class KeysTab(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.threadpool = QtCore.QThreadPool.globalInstance()
        self.app_data = self.parent().app_data
        self.update_private_keys()
        self.update_public_keys()

        self.text_panel = QtWidgets.QTextEdit()
        self.text_panel.setReadOnly(True)

        self.priv_keys_view = QtWidgets.QListView()
        self.priv_keys_view.setModel(self.app_data.priv_keys_model)
        self.priv_keys_view.selectionModel().currentChanged.connect(
            self._update_display
        )

        self.pub_keys_view = QtWidgets.QListView()
        self.pub_keys_view.setModel(self.app_data.pub_keys_model)
        self.pub_keys_view.selectionModel().currentChanged.connect(self._update_display)
        self.pub_keys_view.setSelectionMode(
            QtWidgets.QAbstractItemView.ExtendedSelection
        )

        # When item is selected clear selection in the other list
        self.priv_keys_view.selectionModel().currentChanged.connect(
            lambda: self.pub_keys_view.selectionModel().clear()
        )
        self.pub_keys_view.selectionModel().currentChanged.connect(
            lambda: self.priv_keys_view.selectionModel().clear()
        )

        action_generate_key = Action(
            QtGui.QIcon(":icon/feather/plus-square.png"),
            "Generate new private/public key",
            self,
        )
        action_generate_key.triggered.connect(lambda: KeyGenDialog(parent=self).show())
        action_refresh_keys = QAction(
            QtGui.QIcon(":icon/feather/refresh-cw.png"),
            "Refresh keys from the local keyring",
            self,
        )
        action_refresh_keys.triggered.connect(
            lambda: (
                self.update_private_keys(),
                self.update_public_keys(),
                self._update_display(
                    self.pub_keys_view.selectionModel().currentIndex()
                ),
            )
        )

        toolbar = ToolBar("Key management", self)
        toolbar.addAction(action_generate_key)
        toolbar.addSeparator()
        for action in self.create_public_keys_actions():
            toolbar.addAction(action)
        toolbar.addSeparator()
        toolbar.addAction(action_refresh_keys)

        layout = QtWidgets.QGridLayout()
        layout.addWidget(toolbar, 0, 0, 1, 2)
        layout.addWidget(QtWidgets.QLabel("Private keys"), 1, 0)
        layout.addWidget(self.priv_keys_view, 2, 0)
        layout.addWidget(QtWidgets.QLabel("Public keys"), 1, 1)
        layout.addWidget(self.pub_keys_view, 2, 1)
        layout.addWidget(self.text_panel, 3, 0, 1, 2)
        self.setLayout(layout)

    def key_to_html(self, key: Key) -> str:
        content = []
        content.append("<table>")
        rows = [
            ("User ID", html.escape(str(key.uids[0]))),
            ("Key ID", key.key_id),
            ("Key fingerprint", key.fingerprint),
            ("Key length", key.key_length),
        ]
        for k, v in rows:
            content.append(f"<tr><th>{k}</th><td>{v}</td></tr>")

        content.append("<tr><th>Signatures</th><td>")
        content.append(
            "<br>".join(
                [
                    f"{html.escape(str(sig.issuer_uid))} {sig.issuer_key_id} {sig.signature_class}"
                    for sig in key.valid_signatures
                ]
            )
        )
        content.append("</td></tr>")

        content.append("</table>")
        if key.key_type == crypt.gpg.KeyType.public:
            try:
                crypt.validate_pub_key(
                    key,
                    validation_keyid=self.app_data.config.key_validation_authority_keyid,
                    keyserver_url=self.app_data.config.keyserver_url,
                )
                content.append('<p class="safe">This key has been verified</p>')
            except UserError as e:
                content.append(f'<p class="danger">{e}</p>')
        else:
            content.append(
                "<p>This is a private key. " "Private keys can not be signed.</p>"
            )
        return "".join(content)

    @staticmethod
    def key_to_text(key: Key) -> str:
        return f"User ID: {key.uids[0]}\n" f"Fingerprint: {key.fingerprint}"

    def create_public_keys_actions(self) -> Iterator[QAction]:
        selection_model = self.pub_keys_view.selectionModel()

        def offline_action(icon: str, tip: str, selection: bool = True):
            """Force disable button in offline mode"""
            icon_obj = f":icon/feather/{icon}.png"
            if self.app_data.config.offline:
                action = Action(
                    QtGui.QIcon(icon_obj),
                    f"{tip} (not available in the offline mode)",
                    self,
                )
                action.setEnabled(False)
            else:
                if selection:
                    action = SelectionAction(
                        QtGui.QIcon(icon_obj),
                        tip,
                        self,
                        selection_model=selection_model,
                    )
                else:
                    action = Action(QtGui.QIcon(icon_obj), tip, self)
            return action

        for action, fn in (
            (
                offline_action(
                    "download-cloud",
                    "Search and download keys from the keyserver",
                    False,
                ),
                lambda: KeyDownloadDialog(parent=self).show(),
            ),
            (
                offline_action("upload-cloud", "Upload selected keys to the keyserver"),
                self.upload_key,
            ),
            (
                offline_action("pen-tool", "Request signature for the selected keys"),
                self.send_signature_request,
            ),
            (
                offline_action("rotate-cw", "Update selected keys from the keyserver"),
                self.update_keys,
            ),
            (
                SelectionAction(
                    QtGui.QIcon(":icon/feather/trash-2.png"),
                    "Delete selected keys from your computer",
                    self,
                    selection_model=selection_model,
                ),
                self.delete_keys,
            ),
            (
                Action(
                    QtGui.QIcon(":icon/feather/file-plus.png"),
                    "Import key from file",
                    self,
                ),
                self.import_key,
            ),
        ):
            action.triggered.connect(fn)
            yield action

    def update_private_keys(self):
        # Retrieve all private keys from local keyring.
        keys_private = self.app_data.config.gpg_store.list_sec_keys()
        try:
            default_key = (
                self.app_data.config.default_sender
                or self.app_data.config.gpg_store.default_key()
            )
            self.app_data.default_key_index = next(
                index
                for index, entry in enumerate(keys_private)
                if entry.fingerprint == default_key
            )
            self.app_data.encrypt_sender = default_key
        except StopIteration:
            pass
        self.app_data.priv_keys_model.set_data(
            {f"{key.uids[0]} {key.key_id}": key for key in keys_private}
        )

    def update_public_keys(self):
        # Retrieve all public keys from local keyring.
        keys_public = self.app_data.config.gpg_store.list_pub_keys(sigs=True)
        self.app_data.pub_keys_model.set_data(
            {f"{key.uids[0]} {key.key_id}": key for key in keys_public}
        )

    def update_keys(self):
        """Updated selected keys from the keyserver."""
        show_ok = ok_message("Updated keys", "Keys have been successfully updated.")
        show_warning = warning_callback("GPG key update error")

        selected_keys = [
            index.model().get_value(index).fingerprint
            for index in self.pub_keys_view.selectedIndexes()
        ]
        if selected_keys:
            worker = Worker(
                crypt.download_keys,
                selected_keys,
                self.app_data.config.keyserver_url,
                self.app_data.config.gpg_store,
                report_config=self.app_data.config,
                forward_errors=show_warning,
            )
            worker.signals.result.connect(
                lambda x: (
                    self.update_public_keys(),
                    self._update_display(
                        self.pub_keys_view.selectionModel().currentIndex()
                    ),
                    show_ok(),
                )
            )
            self.threadpool.start(worker)

    def import_key(self):
        path = QtWidgets.QFileDialog.getOpenFileName(
            self, "Select GPG key file", str(Path.home())
        )[0]
        msg = QtWidgets.QMessageBox()
        msg.setWindowTitle("GPG public key import")
        try:
            if path:
                with open(path) as fin:
                    key_data = fin.read()
                crypt.import_keys(key_data, self.app_data.config.gpg_store)
                self.update_public_keys()
                self._update_display(self.pub_keys_view.selectionModel().currentIndex())
                msg.setIcon(QtWidgets.QMessageBox.Information)
                msg.setText("Key has been imported.")
                msg.exec_()
        except (UnicodeDecodeError, UserError) as e:
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText(format(e))
            msg.exec_()

    def delete_keys(self):
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Question)
        msg.setWindowTitle("Delete public key")
        msg.setText("Do you really want to delete the following public key?")
        msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
        msg_warn = QtWidgets.QMessageBox()
        msg_warn.setWindowTitle("GPG key deletion error")
        msg_warn.setIcon(QtWidgets.QMessageBox.Warning)
        priv_keys = self.app_data.config.gpg_store.list_sec_keys()

        selected_keys = [
            index.model().get_value(index)
            for index in self.pub_keys_view.selectedIndexes()
        ]
        for key in selected_keys:
            if any(k for k in priv_keys if key.fingerprint == k.fingerprint):
                msg_warn.setText(
                    "Unable to delete key:\n\n"
                    f"{key.uids[0]}\n{key.fingerprint}\n\n"
                    "Deleting private keys (and by extension public keys "
                    "with an associated private key) is not supported by "
                    "this application. Please use an external software  "
                    "such as GnuPG (Linux, MacOS) or Kleopatra (Windows)."
                )
                msg_warn.exec_()
                continue
            msg.setDetailedText(self.key_to_text(key))
            if key is selected_keys[0]:
                click_show_details(msgbox=msg)
            status = msg.exec_()
            if status == QtWidgets.QMessageBox.Ok:
                try:
                    crypt.delete_keys([key.fingerprint], self.app_data.config.gpg_store)
                    self.pub_keys_view.selectionModel().clearSelection()
                except UserError as e:
                    msg_warn.setText(format(e))
                    msg_warn.exec_()
                self.text_panel.clear()
        self.update_public_keys()

    def upload_key(self):
        msg = QtWidgets.QMessageBox()
        msg.setWindowTitle("Send public key")
        msg.setText("Do you want to upload the selected key to the key server?")
        msg.setIcon(QtWidgets.QMessageBox.Question)
        msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
        show_ok = ok_message(
            "Send public key", "Key has been successfully uploaded to the keyserver."
        )
        show_warning = warning_callback("GPG key upload error")

        selected_keys = [
            index.model().get_value(index)
            for index in self.pub_keys_view.selectedIndexes()
        ]
        for key in selected_keys:
            msg.setDetailedText(self.key_to_text(key))
            if key is selected_keys[0]:
                click_show_details(msgbox=msg)
            status = msg.exec_()
            if status == QtWidgets.QMessageBox.Ok:
                worker = Worker(
                    upload_keys_workflow.upload_keys,
                    [key.fingerprint],
                    config=self.app_data.config,
                    report_config=self.app_data.config,
                    capture_loggers=(upload_keys_workflow.logger,),
                    forward_errors=show_warning,
                )
                worker.signals.result.connect(lambda _: show_ok())
                self.threadpool.start(worker)

    def send_signature_request(self):
        selected_keys = [
            index.model().get_value(index)
            for index in self.pub_keys_view.selectedIndexes()
        ]
        request_signature(selected_keys, self.app_data.config, self.threadpool, self)

    def _update_display(self, index: QtCore.QModelIndex):
        style = (
            "<style>"
            "th {text-align: left; padding: 0 20px 5px 0;}"
            ".danger { color: red;}"
            ".safe { color: green;}"
            "</style>"
        )
        if index.isValid():
            try:
                self.text_panel.setHtml(
                    style + self.key_to_html(index.model().get_value(index))
                )
            except IndexError:
                self.text_panel.setHtml("")


class KeyDownloadDialog(QtWidgets.QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.threadpool = QtCore.QThreadPool.globalInstance()
        self.setWindowTitle("Download public keys from the server")
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)

        self.search_string = LineEdit()
        self.btn_search = QtWidgets.QPushButton("Search")
        self.btn_search.clicked.connect(self.search_keys)

        self.key_list_view = QtWidgets.QListView()
        key_list_model = QtCore.QStringListModel()
        self.key_list_view.setModel(key_list_model)

        self.btn_download = QtWidgets.QPushButton("Download")
        self.btn_download.clicked.connect(self.download_selected)

        btn_cancel = QtWidgets.QPushButton("Close")
        btn_cancel.clicked.connect(self.close)

        layout = QtWidgets.QGridLayout()
        layout.addWidget(
            QtWidgets.QLabel(
                "Enter a search term (e.g. user name, email, key " "fingerprint)"
            ),
            0,
            0,
        )
        layout.addWidget(self.search_string, 1, 0)
        layout.addWidget(self.btn_search, 1, 1)
        layout.addWidget(QtWidgets.QLabel("Select a key to download"), 2, 0)
        layout.addWidget(self.key_list_view, 3, 0)
        layout.addWidget(self.btn_download, 3, 1)
        layout.addWidget(btn_cancel, 4, 1)
        self.setLayout(layout)

    def search_keys(self):
        self.btn_search.setEnabled(False)
        self.key_list_view.model().setStringList([])
        worker = Worker(
            crypt.search_keyserver,
            self.search_string.text(),
            self.parent().app_data.config.keyserver_url,
            report_config=self.parent().app_data.config,
        )
        worker.signals.result.connect(
            lambda keys: self.key_list_view.model().setStringList(
                [f"{k.uid} {k.fingerprint}" for k in keys]
            )
        )
        worker.signals.finished.connect(lambda: self.btn_search.setEnabled(True))
        self.threadpool.start(worker)

    def download_selected(self):
        key_ids = []
        for index in self.key_list_view.selectedIndexes():
            key_ids.append(index.model().data(index).split()[-1])
        if key_ids:
            self.btn_download.setEnabled(False)
            worker = Worker(
                crypt.download_keys,
                key_ids,
                self.parent().app_data.config.keyserver_url,
                self.parent().app_data.config.gpg_store,
                report_config=self.parent().app_data.config,
            )
            worker.signals.result.connect(lambda x: self.parent().update_public_keys())
            worker.signals.finished.connect(lambda: self.btn_download.setEnabled(True))
            self.threadpool.start(worker)


class KeyGenDialog(QtWidgets.QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.threadpool = QtCore.QThreadPool.globalInstance()
        self.setWindowTitle("Generate new key pair")
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)

        self.text_name_full = LineEdit()
        self.text_name_extra = LineEdit()
        self.text_email = LineEdit()
        self.text_pass = LineEdit()
        self.text_pass_repeat = LineEdit()
        self.toggle_password_visibility(False)

        re_email = QtCore.QRegularExpression(r"[^@]+@[^@]+\.[^@]+")
        self.text_email.setValidator(QtGui.QRegularExpressionValidator(re_email))

        self.btn_run = QtWidgets.QPushButton("Generate key")
        self.btn_run.setDefault(True)
        self.btn_run.clicked.connect(self.create_private_key)
        btn_cancel = QtWidgets.QPushButton("Close")
        btn_cancel.clicked.connect(self.close)
        btn_show_pass = QtWidgets.QPushButton("Show")
        btn_show_pass.setCheckable(True)
        btn_show_pass.clicked.connect(self.toggle_password_visibility)

        layout = QtWidgets.QGridLayout()
        layout.addWidget(QtWidgets.QLabel("Full name"), 0, 0)
        layout.addWidget(self.text_name_full, 0, 1)
        layout.addWidget(QtWidgets.QLabel("(optional) institution/project"), 1, 0)
        layout.addWidget(self.text_name_extra, 1, 1)
        layout.addWidget(QtWidgets.QLabel("Institutional email"), 2, 0)
        layout.addWidget(self.text_email, 2, 1)
        layout.addWidget(QtWidgets.QLabel("Password"), 3, 0)
        layout.addWidget(self.text_pass, 3, 1)
        layout.addWidget(QtWidgets.QLabel("Password (repeat)"), 4, 0)
        layout.addWidget(self.text_pass_repeat, 4, 1)
        layout.addWidget(btn_show_pass, 4, 2)
        layout.addWidget(btn_cancel, 5, 0)
        layout.addWidget(self.btn_run, 5, 1)
        layout.addWidget(
            QtWidgets.QLabel("Key generation can take a few minutes"), 6, 0, 1, 3
        )
        self.setLayout(layout)

    def toggle_password_visibility(self, show: bool):
        mode = QtWidgets.QLineEdit.Normal if show else QtWidgets.QLineEdit.Password
        self.text_pass.setEchoMode(mode)
        self.text_pass_repeat.setEchoMode(mode)

    def clear_form(self):
        self.text_name_full.clear()
        self.text_name_extra.clear()
        self.text_email.clear()
        self.text_pass.clear()
        self.text_pass_repeat.clear()

    def post_key_creation(self, key: Key):
        msg = QtWidgets.QMessageBox(self)
        msg.setWindowTitle("GPG Key Generation")
        msg.setIcon(QtWidgets.QMessageBox.Information)
        try:
            revocation_cert = crypt.create_revocation_certificate(
                key.fingerprint,
                self.text_pass.text(),
                self.parent().app_data.config.gpg_store,
            )
            msg.setText(
                "Your new key has been successfully generated.\n\n"
                "Additionally, a revocation certificate was also created. "
                "It can be used to revoke your key in the eventuality that it "
                "gets compromised, lost, or that you forgot your password.\n"
                "Please store the revocation certificate below in a safe "
                "location, as anyone can use it to revoke your key."
            )
            msg.setDetailedText(revocation_cert.decode())
            # Programatically click the "Show Details..." button so that the
            # certificate is shown by default.
            click_show_details(msgbox=msg)

        except UserError:
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText(
                "Key has been successfully generated. "
                "However, it was not possible to create a revocation "
                "certificate. "
                "Execute the following command to create the certificate\n\n"
                f"gpg --gen-revoke {key.fingerprint}"
            )
        finally:
            msg.exec_()
            if (
                self.parent().app_data.config.key_validation_authority_keyid
                and not self.parent().app_data.config.offline
            ):
                request_signature(
                    (key,), self.parent().app_data.config, self.threadpool, self
                )
            self.clear_form()
            self.parent().update_private_keys()
            self.parent().update_public_keys()
            self.close()

    def create_private_key(self):
        show_warning = warning_callback("GPG Key Generation Error")
        self.btn_run.setEnabled(False)
        full_name = f"{self.text_name_full.text()} " f"{self.text_name_extra.text()}"
        worker = Worker(
            crypt.create_key,
            full_name,
            self.text_email.text(),
            self.text_pass.text(),
            self.text_pass_repeat.text(),
            self.parent().app_data.config.gpg_store,
            forward_errors=show_warning,
            report_config=self.parent().app_data.config,
        )
        worker.signals.result.connect(self.post_key_creation)
        worker.signals.finished.connect(lambda: self.btn_run.setEnabled(True))
        self.threadpool.start(worker)


def click_show_details(msgbox: QtWidgets.QMessageBox):
    for button in msgbox.buttons():
        if msgbox.buttonRole(button) is QtWidgets.QMessageBox.ButtonRole.ActionRole:
            button.click()


def request_signature(
    keys: Sequence[Key],
    config: Config,
    threadpool: QtCore.QThreadPool,
    parent: Optional[QtWidgets.QWidget] = None,
):
    msg = QtWidgets.QMessageBox(parent)
    msg.setWindowTitle("Key signing request")
    msg.setText("Do you want to request signature for this key?")
    msg.setIcon(QtWidgets.QMessageBox.Question)
    msg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
    show_ok = ok_message(
        "Key signing request", "Key signing request has been sent.", parent=parent
    )
    show_warning = warning_callback("Key signing request", parent=parent)

    for key in keys:
        msg.setDetailedText(KeysTab.key_to_text(key))
        if key is keys[0]:
            click_show_details(msgbox=msg)
        status = msg.exec_()
        if status == QtWidgets.QMessageBox.Yes:
            worker = Worker(
                request_sigs_workflow.request_sigs,
                key_ids=[key.key_id],
                config=config,
                capture_loggers=(request_sigs_workflow.logger,),
                forward_errors=show_warning,
                report_config=config,
            )
            worker.signals.result.connect(lambda _: show_ok())
            threadpool.start(worker)


def ok_message(title: str, msg: str, parent=None) -> Callable:
    msg_ok = QtWidgets.QMessageBox(parent)
    msg_ok.setIcon(QtWidgets.QMessageBox.Information)
    msg_ok.setWindowTitle(title)
    msg_ok.setText(msg)
    return msg_ok.exec_



#!/usr/bin/env python3
"""Secure Encryption and Transfer Tool."""

import logging
import sys

from sett.gui.component import is_macos, MACOS_TOOLTIP_BG_COLOR
from . import main_window
from .pyside import QtWidgets, QtGui
from .resources import rc_icons
from ..utils.log import log_to_rotating_file


def run():
    application = QtWidgets.QApplication(sys.argv)
    if is_macos():
        palette = application.palette()
        palette.setColor(QtGui.QPalette.ColorRole.ToolTipBase, MACOS_TOOLTIP_BG_COLOR)
        application.setPalette(palette)
    window = main_window.MainWindow()
    log_to_rotating_file(
        log_dir=window.app_data.config.log_dir,
        file_max_number=window.app_data.config.log_max_file_number,
    )
    window.show()
    return application.exec_()


if __name__ == "__main__":
    run()



from functools import partial
from pathlib import Path
from typing import Callable, Optional, Tuple

from .model import AppData
from .parallel import Worker
from .pyside import QtCore, QtWidgets, QAction, QtGui
from ..utils.progress import ProgressInterface

EXTRA_HEIGHT = 1
MACOS_TOOLTIP_BG_COLOR = "#ffffca"


def is_macos():
    return QtCore.QSysInfo.productType() in ("osx", "macos")


class Action(QAction):
    def __init__(
        self,
        icon: QtGui.QIcon,
        text: str,
        parent: Optional[QtWidgets.QWidget] = None,
    ):
        super().__init__(icon, text, parent)
        if is_macos():
            self.setToolTip(
                f"<p style='background-color: {MACOS_TOOLTIP_BG_COLOR};'>{text}</p>"
            )


class SpinBox(QtWidgets.QSpinBox):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)
        if is_macos():
            self.setMinimumHeight(self.sizeHint().height() + EXTRA_HEIGHT)


class LineEdit(QtWidgets.QLineEdit):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)
        if is_macos():
            self.setMinimumHeight(self.sizeHint().height() + EXTRA_HEIGHT)


class ToolBar(QtWidgets.QToolBar):
    def __init__(self, title: str, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(title, parent)
        if is_macos():
            self.setStyleSheet("background: transparent;border-width: 0px;")
        self.setIconSize(QtCore.QSize(20, 20))


class MandatoryLabel(QtWidgets.QLabel):
    """A label extension which appends a '*' to label's end to mark the field as required."""

    def __init__(self, label: str):
        super().__init__(label + " <sup><font color='red'>*</font></sup>")


class SelectionButton(QtWidgets.QPushButton):
    """A push button extension which connects this button to given selection model.

    The button is disabled by default. And gets enabled when the selection has, at least,
    one selected row.
    """

    def __init__(self, label: str, selection_model: QtCore.QItemSelectionModel):
        super().__init__(label)
        self.setEnabled(False)
        self._selection_model = selection_model
        selection_model.selectionChanged.connect(self.selection_changed)

    def selection_changed(self):
        # Following works better than using 'QItemSelection', especially in cases
        # where multiple selection is possible
        self.setEnabled(bool(len(self._selection_model.selectedRows())))


class SelectionAction(Action):
    """An action extension which connects this action to given selection model.

    The action is disabled by default. And gets enabled when the selection has, at least,
    one selected row.
    """

    def __init__(self, *args, selection_model: QtCore.QItemSelectionModel):
        super().__init__(*args)
        self.setEnabled(False)
        self._selection_model = selection_model
        selection_model.selectionChanged.connect(self.selection_changed)

    def selection_changed(self):
        # Following works better than using 'QItemSelection', especially in cases
        # where multiple selection is possible
        self.setEnabled(bool(len(self._selection_model.selectedRows())))


class GuiProgress(QtCore.QObject, ProgressInterface):
    updated = QtCore.Signal(int)

    def __init__(self):
        super().__init__()
        self.n = 0

    def update(self, completed_fraction):
        self.n = completed_fraction
        self.updated.emit(round(completed_fraction * 100, 0))

    def get_completed_fraction(self):
        return self.n


class ConsoleWidget(QtWidgets.QGroupBox):
    def __init__(self, title, parent):
        super().__init__(title, parent)
        self.textbox = QtWidgets.QPlainTextEdit()
        self.textbox.setReadOnly(True)
        btn_clear_console = QtWidgets.QPushButton(
            QtGui.QIcon(":icon/feather/slash.png"), ""
        )
        btn_clear_console.setToolTip("Clear console")
        btn_clear_console.clicked.connect(self.clear)
        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(self.textbox)
        layout.addWidget(btn_clear_console, alignment=QtCore.Qt.AlignTop)

    def clear(self) -> None:
        self.textbox.clear()

    def append(self, text: str) -> None:
        self.textbox.appendPlainText(text)

    def write(self, text: str) -> None:
        self.textbox.appendPlainText(text)


class PathInput:
    """Path selection widget with a select button and a show path field."""

    def __init__(self, directory=True, path: Optional[Path] = Path.home(), parent=None):
        self.parent = parent
        self.text = LineEdit(parent)
        self.text.setReadOnly(True)
        self.btn = QtWidgets.QPushButton(
            QtGui.QIcon(f":icon/feather/{'folder' if directory else 'file'}.png"),
            "",
        )
        self.btn.setToolTip("Change location")
        self.btn.clicked.connect(partial(self._update_location, directory))
        # Additional action to clear the selected path
        self.btn_clear = QtWidgets.QPushButton(
            QtGui.QIcon(":icon/feather/slash.png"), ""
        )
        self.btn_clear.setToolTip("Clear location")
        self.btn_clear.clicked.connect(self._clear_location)
        self.update_path(path)

    def update_path(self, path: Optional[Path]):
        self.path = path
        self.text.setText("" if path is None else str(path))
        self.text.editingFinished.emit()

    def _update_location(self, directory: bool):
        if self.path and self.path.exists():
            location = self.path if self.path.is_dir() else self.path.parent
        else:
            location = Path.home()
        if directory:
            new_path = QtWidgets.QFileDialog.getExistingDirectory(
                self.parent, "Select Directory", str(location)
            )
        else:
            new_path, _ = QtWidgets.QFileDialog.getOpenFileName(
                self.parent, "Select File", str(location)
            )
        if new_path:
            self.update_path(Path(new_path))

    def _clear_location(self):
        self.update_path(None)

    @property
    def status_tip(self) -> str:
        return self.text.statusTip()

    @status_tip.setter
    def status_tip(self, msg: str):
        self.text.setStatusTip(msg)

    def on_path_change(self, fn: Callable[[Optional[Path]], None]) -> None:
        """Run callback when path changes."""
        self.text.editingFinished.connect(lambda: fn(self.path))


class TabMixin:
    def create_console(self):
        self.console = ConsoleWidget("Console", self)

    def create_progress_bar(self):
        self.progress_bar = QtWidgets.QProgressBar(self)

    @staticmethod
    def _create_disabled_button(action_name: str) -> QtWidgets.QPushButton:
        button = QtWidgets.QPushButton(action_name)
        button.setEnabled(False)
        return button

    def set_buttons_enabled(self, enabled: bool):
        self.btn_run.setEnabled(enabled)
        self.btn_test.setEnabled(enabled)

    def create_run_panel(self, panel_name: str, action: Callable, action_name: str):
        self.run_panel = QtWidgets.QGroupBox(panel_name)
        layout = QtWidgets.QHBoxLayout(self.run_panel)
        self.btn_test = TabMixin._create_disabled_button("Test")
        # On pressed button, make sure that the focus switches to that button (Mac specific issue)
        self.btn_test.pressed.connect(self.btn_test.setFocus)
        self.btn_test.clicked.connect(partial(action, dry_run=True))
        self.btn_run = TabMixin._create_disabled_button(action_name)
        # On pressed button, make sure that the focus switches to that button (Mac specific issue)
        self.btn_run.pressed.connect(self.btn_run.setFocus)
        self.btn_run.clicked.connect(action)
        layout.addWidget(self.btn_test)
        layout.addWidget(self.btn_run)

    def add_worker_actions(self, worker: Worker):
        """Attach GUI-updating signals to worker"""
        worker.signals.started.connect(lambda: self.run_panel.setEnabled(False))
        worker.signals.logging.connect(self.console.write)
        worker.signals.error.connect(lambda e: self.console.append(str(e[1])))
        worker.signals.finished.connect(lambda: self.run_panel.setEnabled(True))


def run_dialog(parent, msg: str, password: bool = True):
    dialog_pwd = QtWidgets.QInputDialog(parent)
    dialog_pwd.setLabelText(msg)
    dialog_pwd.setWindowTitle("SETT")
    if password:
        dialog_pwd.setTextEchoMode(QtWidgets.QLineEdit.EchoMode.Password)
    if dialog_pwd.exec_() != QtWidgets.QDialog.Accepted or not dialog_pwd.textValue():
        return None
    return dialog_pwd.textValue()


def warning_callback(
    title: str, parent: Optional[QtWidgets.QWidget] = None
) -> Callable:
    msg_warn = warning_dialog(title, parent=parent)

    def _show_warning(msg: str):
        msg_warn.setText(format(msg))
        msg_warn.exec_()

    return _show_warning


def warning_dialog(
    title: str, text: Optional[str] = None, parent: Optional[QtWidgets.QWidget] = None
) -> QtWidgets.QMessageBox:
    msg = QtWidgets.QMessageBox(parent=parent)
    msg.setWindowTitle(title)
    msg.setIcon(QtWidgets.QMessageBox.Warning)
    if text is not None:
        msg.setText(text)
    return msg


def show_warning(title: str, text: str, parent: Optional[QtWidgets.QWidget] = None):
    warning_dialog(title, text, parent).show()


def create_verify_dtr_checkbox(
    app_data: AppData, field_name: str, parent: Optional[QtWidgets.QWidget] = None
) -> QtWidgets.QCheckBox:
    verify_dtr = QtWidgets.QCheckBox("Verify DTR ID", parent)
    verify_dtr.setStatusTip("Verify DTR (Data Transfer Request) ID")
    verify_dtr.setChecked(not app_data.config.offline and app_data.encrypt_verify_dtr)
    if app_data.config.offline:
        verify_dtr.setEnabled(not app_data.config.offline)
        verify_dtr.setStatusTip(
            "Verifying Transfer Request ID in offline mode is not possible"
        )
    verify_dtr.stateChanged.connect(
        lambda state: setattr(app_data, field_name, state == QtCore.Qt.Checked)
    )
    return verify_dtr


class FieldMapping:
    def __init__(self):
        self._mapping = {}

    def bind_text(self, field_name, w):
        self._mapping[field_name] = w.setText
        return w

    def bind_checked(self, field_name, w):
        self._mapping[field_name] = w.setChecked
        return w

    def bind_value(self, field_name, w):
        self._mapping[field_name] = w.setValue
        return w

    def bind_path(self, field_name, w):
        self._mapping[field_name] = lambda p: w.update_path(p and Path(p))
        return w

    def load(self, parameters):
        for key, val in parameters.items():
            try:
                self._mapping[key](val)
            except KeyError:
                raise ValueError(
                    f"Invalid field `{key}` found in your config file"
                ) from None
        for key in set(self._mapping.keys()) - set(parameters.keys()):
            self._mapping[key](None)


def create_slider(
    minimum: int,
    maximum: int,
    initial_value: int,
    status_tip: Optional[str],
    on_change: Optional[Callable[[int], None]] = None,
    show_ticks: bool = False,
    interval: int = 1,
) -> Tuple[QtWidgets.QSlider, QtWidgets.QLabel]:
    widget_value = QtWidgets.QLabel(str(initial_value))

    slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
    slider.setMinimum(minimum)
    slider.setMaximum(maximum)
    slider.setValue(initial_value)
    slider.setTickInterval(interval)
    if status_tip is not None:
        slider.setStatusTip(status_tip)
    if show_ticks:
        slider.setTickPosition(QtWidgets.QSlider.TicksBelow)

    def update_text(value: int):
        widget_value.setText(str(value))

    slider.valueChanged.connect(update_text)
    if on_change is not None:
        slider.valueChanged.connect(on_change)

    return slider, widget_value



import warnings
import re
import urllib.error

from sett import URL_READTHEDOCS, APP_NAME_SHORT
from ..core.request import urlopen
from .. import __version__, __project_name__


def check_version(repo_url: str, gui_formatting: bool = False):
    latest = get_latest_version(repo_url)
    if (
        not __version__.startswith("0.0.0.dev")
        and __version__ != latest
        and latest is not None
    ):
        line_break = "\n"
        update_doc_url = (
            f"{URL_READTHEDOCS}/en/stable/" "sett_installation.html#updating-sett"
        )
        if gui_formatting:
            line_break = "<br>"
            update_doc_url = f"<a href='{update_doc_url}'>{update_doc_url}</a>"
        warnings.warn(
            f"Your {APP_NAME_SHORT} version ({__version__}) is "
            f"outdated and no longer supported.{line_break}"
            f"Please upgrade to the latest version ({latest}) as "
            f"soon as possible.{line_break * 2}"
            f"In most cases, {APP_NAME_SHORT} can be upgraded with "
            f"the following commmand:{line_break}"
            f"pip install --upgrade --user sett{line_break * 2}"
            f"Documentation on how to upgrade your {APP_NAME_SHORT} "
            f"installation can also be found at:{line_break}"
            f"{update_doc_url}{line_break}"
        )


def get_latest_version(repo_url: str):
    try:
        url = repo_url + "/simple/" + __project_name__ + "/"
        with urlopen(url) as response:
            versions = re.findall(
                __project_name__.encode() + b"-([0-9]*.[0-9]*.[0-9]*).tar.gz",
                response.read(),
            )
            version = max(tuple(map(int, v.split(b"."))) for v in versions)
    except urllib.error.URLError:
        warnings.warn(
            "Could not connect to pypi repository to query the latest version. "
            "You might have an outdated version. Please check yourself!"
        )
    except IndexError:
        warnings.warn(
            "No releases found on the pypi repository! "
            "Please contact the developers!"
        )
    return ".".join(map(str, version))



import enum
import operator
import time
from functools import partial
from typing import Iterator, Optional, Sequence

from .component import (
    TabMixin,
    PathInput,
    GuiProgress,
    run_dialog,
    MandatoryLabel,
    create_verify_dtr_checkbox,
    FieldMapping,
    ToolBar,
    LineEdit,
    Action,
)
from .file_selection_widget import ArchiveOnlyFileSelectionWidget
from .parallel import Worker
from .pyside import QtCore, QtGui, QtWidgets, QAction
from .. import protocols
from ..core.error import UserError
from ..utils.config import Connection, ConnectionStore
from ..workflows import transfer


class TransferTab(QtWidgets.QWidget, TabMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.threadpool = QtCore.QThreadPool.globalInstance()
        self.app_data = self.parent().app_data
        self.app_data.transfer_protocol_args = {"sftp": {}, "liquid_files": {}}
        self.connections_model = QtCore.QStringListModel()

        files_panel = self.create_files_panel()

        verify_dtr_group = QtWidgets.QGroupBox()
        verify_dtr_layout = QtWidgets.QVBoxLayout(verify_dtr_group)
        verify_dtr_layout.addWidget(
            create_verify_dtr_checkbox(self.app_data, "transfer_verify_dtr")
        )

        options_panel = self.create_options_panel()
        self.create_run_panel("Transfer data", self.transfer, "Transfer selected files")
        self.app_data.add_listener("transfer_files", self._enable_buttons)
        self.create_console()
        self.create_progress_bar()

        layout = QtWidgets.QVBoxLayout(self)
        for widget in (
            files_panel,
            verify_dtr_group,
            options_panel,
            self.run_panel,
            self.console,
            self.progress_bar,
        ):
            layout.addWidget(widget)

    def _enable_buttons(self):
        self.set_buttons_enabled(len(self.app_data.transfer_files) > 0)

    def create_files_panel(self):
        box = ArchiveOnlyFileSelectionWidget(
            "Encrypted files to transfer (hint: you can drag & drop files)", self
        )
        box.file_list_model.layoutChanged.connect(
            lambda: setattr(self.app_data, "transfer_files", box.get_list())
        )
        return box

    def create_sftp_options_panel(self):
        field_mapping = FieldMapping()

        text_username = field_mapping.bind_text("username", LineEdit())
        text_username.setStatusTip("Username on the SFTP server")
        text_destination_dir = field_mapping.bind_text("destination_dir", LineEdit())
        text_username.textChanged.connect(
            lambda: operator.setitem(
                self.app_data.transfer_protocol_args["sftp"],
                "username",
                text_username.text(),
            )
        )
        text_host = field_mapping.bind_text("host", LineEdit())
        text_host.setStatusTip("URL of the SFTP server with an optional port " "number")
        text_host.textChanged.connect(
            lambda: operator.setitem(
                self.app_data.transfer_protocol_args["sftp"], "host", text_host.text()
            )
        )
        text_jumphost = field_mapping.bind_text("jumphost", LineEdit())
        text_jumphost.setStatusTip("(optional) URL of the jumphost server")
        text_jumphost.textChanged.connect(
            lambda: operator.setitem(
                self.app_data.transfer_protocol_args["sftp"],
                "jumphost",
                text_jumphost.text() or None,
            )
        )
        text_destination_dir.textChanged.connect(
            lambda: operator.setitem(
                self.app_data.transfer_protocol_args["sftp"],
                "destination_dir",
                text_destination_dir.text(),
            )
        )
        text_destination_dir.setStatusTip(
            "Relative or absolute path to the "
            "destination directory on the SFTP "
            "server"
        )
        pkey_location = field_mapping.bind_path(
            "pkey", PathInput(directory=False, path=None)
        )
        pkey_location.status_tip = "Path to the private SSH key used for authentication"
        pkey_location.on_path_change(
            lambda path: operator.setitem(
                self.app_data.transfer_protocol_args["sftp"],
                "pkey",
                "" if path is None else str(path),
            )
        )
        pkey_password = field_mapping.bind_text("pkey_password", LineEdit())
        pkey_password.setStatusTip("Passphrase for the SSH private key")
        pkey_password.setEchoMode(QtWidgets.QLineEdit.Password)
        pkey_password.textChanged.connect(
            lambda: operator.setitem(
                self.app_data.transfer_protocol_args["sftp"],
                "pkey_password",
                pkey_password.text(),
            )
        )

        box = QtWidgets.QGroupBox()
        box.setFlat(True)
        box.load_connection = field_mapping.load
        layout = QtWidgets.QGridLayout(box)
        layout.addWidget(MandatoryLabel("User name"), 0, 0)
        layout.addWidget(text_username, 0, 1)
        layout.addWidget(MandatoryLabel("Host URL"), 1, 0)
        layout.addWidget(text_host, 1, 1)
        layout.addWidget(QtWidgets.QLabel("Jumphost URL"), 2, 0)
        layout.addWidget(text_jumphost, 2, 1)
        layout.addWidget(MandatoryLabel("Destination directory"), 3, 0)
        layout.addWidget(text_destination_dir, 3, 1)
        layout.addWidget(QtWidgets.QLabel("SSH key location"), 4, 0)
        layout.addWidget(pkey_location.text, 4, 1)
        layout.addWidget(pkey_location.btn, 4, 2)
        layout.addWidget(pkey_location.btn_clear, 4, 3)
        layout.addWidget(QtWidgets.QLabel("SSH key password"), 5, 0)
        layout.addWidget(pkey_password, 5, 1)
        return box

    def create_liquidfiles_options_panel(self):
        field_mapping = FieldMapping()
        text_host = field_mapping.bind_text("host", LineEdit())
        text_api_key = field_mapping.bind_text("api_key", LineEdit())
        text_host.textChanged.connect(
            lambda: operator.setitem(
                self.app_data.transfer_protocol_args["liquid_files"],
                "host",
                text_host.text(),
            )
        )
        text_api_key.textChanged.connect(
            lambda: operator.setitem(
                self.app_data.transfer_protocol_args["liquid_files"],
                "api_key",
                text_api_key.text(),
            )
        )
        box = QtWidgets.QGroupBox()
        box.setFlat(True)
        box.load_connection = field_mapping.load
        layout = QtWidgets.QGridLayout(box)
        layout.addWidget(MandatoryLabel("Host URL"), 0, 0)
        layout.addWidget(text_host, 0, 1)
        layout.addWidget(MandatoryLabel("API Key"), 1, 0)
        layout.addWidget(text_api_key, 1, 1)
        return box

    def create_conn_actions(
        self, connections_selection: QtWidgets.QComboBox
    ) -> Iterator[QAction]:
        connection_store = ConnectionStore()

        def confirm(text: str) -> bool:
            dialog = QtWidgets.QMessageBox(parent=self)
            dialog.setWindowTitle("Connection profile")
            dialog.setText(text)
            dialog.setStandardButtons(
                QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel
            )
            dialog.setDefaultButton(QtWidgets.QMessageBox.Cancel)
            return dialog.exec_() == QtWidgets.QMessageBox.Ok

        def rename():
            label_old = connections_selection.currentText()
            dialog = ConnectionDialog(
                parent=self,
                label=f"Rename connection profile '{label_old}' to:",
                forbidden=self.connections_model.stringList(),
                default_input=label_old,
            )
            if dialog.exec_():
                label_new = dialog.text_field.text()
                self.app_data.config.connections[
                    label_new
                ] = self.app_data.config.connections.pop(label_old)
                try:
                    connection_store.rename(old=label_old, new=label_new)
                except UserError:
                    # Connection profile is not saved in the config file
                    pass
                self.connections_model.setStringList(
                    sorted(
                        label_new if x == label_old else x
                        for x in self.connections_model.stringList()
                    )
                )
                connections_selection.setCurrentText(label_new)

        def add():
            dialog = ConnectionDialog(
                parent=self,
                label="New connection profile label:",
                forbidden=self.connections_model.stringList(),
            )
            if dialog.exec_():
                label = dialog.text_field.text()
                self.app_data.config.connections[label] = Connection(
                    protocol="sftp", parameters={}
                )
                self.connections_model.setStringList(
                    sorted(self.connections_model.stringList() + [label])
                )
                connections_selection.setCurrentText(label)

        def delete():
            label = connections_selection.currentText()
            if confirm(f"Do you want to delete '{label}' profile?"):
                self.app_data.config.connections.pop(label)
                self.connections_model.removeRow(connections_selection.currentIndex())
                try:
                    connection_store.delete(label)
                except UserError:
                    # Connection profile is not saved in the config file
                    pass

        def save():
            label = connections_selection.currentText()
            if confirm(f"Do you want to save '{label}' profile?"):
                protocol_name = self.app_data.transfer_protocol_name
                connection = Connection(
                    protocol=protocol_name,
                    parameters=self.app_data.transfer_protocol_args[
                        protocol_name
                    ].copy(),
                )
                self.app_data.config.connections[label] = connection
                connection_store.save(label, connection)

        def set_btn_enabled(btn: QtWidgets.QAbstractButton, text: str):
            btn.setEnabled(bool(text))

        for tip, fn, needs_selection, icon in (
            ("Create a new connection profile", add, False, "plus-square"),
            ("Rename the current connection profile", rename, True, "edit"),
            ("Delete the current connection profile", delete, True, "trash-2"),
            ("Save the current connection profile", save, True, "save"),
        ):
            action = Action(QtGui.QIcon(f":icon/feather/{icon}.png"), tip, self)
            action.triggered.connect(fn)
            if needs_selection:
                connections_selection.currentTextChanged.connect(
                    partial(set_btn_enabled, action)
                )
                if not connections_selection.currentText():
                    action.setEnabled(False)
            yield action

    def create_options_panel(self):
        box = QtWidgets.QGroupBox("Connection")

        connections_selection = QtWidgets.QComboBox(box)
        connections_selection.setStatusTip(
            "Select a predefined connection profile. Check documentation for details."
        )
        connections_selection.setModel(self.connections_model)
        self.connections_model.setStringList(list(self.app_data.config.connections))

        protocol_btn_grp = QtWidgets.QButtonGroup(box)
        protocol_btn_grp.addButton(QtWidgets.QRadioButton("sftp"))
        protocol_btn_grp.addButton(QtWidgets.QRadioButton("liquid_files"))
        get_btn_from_group(protocol_btn_grp, "sftp").setChecked(True)

        protocol_boxes = {
            "sftp": self.create_sftp_options_panel(),
            "liquid_files": self.create_liquidfiles_options_panel(),
        }
        protocol_boxes["liquid_files"].hide()

        def load_connection():
            connection = self.app_data.config.connections.get(
                connections_selection.currentText()
            )
            if not connection:
                return
            get_btn_from_group(protocol_btn_grp, connection.protocol).click()
            protocol_boxes[connection.protocol].load_connection(connection.parameters)
            self.app_data.transfer_protocol_args[
                self.app_data.transfer_protocol_name
            ].update(connection.parameters)

        connections_selection.currentTextChanged.connect(load_connection)

        def toggle_protocol(btn, state: bool):
            if state:
                self.app_data.transfer_protocol_name = btn.text()
            protocol_boxes[btn.text()].setVisible(state)

        protocol_btn_grp.buttonToggled.connect(toggle_protocol)

        toolbar = ToolBar("Options", self)
        for action in self.create_conn_actions(connections_selection):
            toolbar.addAction(action)
        layout_connection = QtWidgets.QHBoxLayout()
        layout_connection.addWidget(connections_selection)
        layout_connection.addWidget(toolbar)

        layout_protocol_buttons = QtWidgets.QHBoxLayout()
        for btn in protocol_btn_grp.buttons():
            layout_protocol_buttons.addWidget(btn)

        layout_protocol = QtWidgets.QVBoxLayout()
        layout_protocol.addLayout(layout_protocol_buttons)
        for protocol_box in protocol_boxes.values():
            layout_protocol.addWidget(protocol_box)

        layout = QtWidgets.QVBoxLayout(box)
        layout.addLayout(layout_connection)
        layout.addLayout(layout_protocol)
        load_connection()
        return box

    def transfer(self, dry_run=False):
        progress = GuiProgress()
        progress.updated.connect(self.progress_bar.setValue)

        second_factor = None

        class Msg(enum.Enum):
            code = enum.auto()

        class MsgSignal(QtCore.QObject):
            msg = QtCore.Signal(object)

        msg_signal = MsgSignal()

        def second_factor_callback():
            msg_signal.msg.emit(Msg.code)
            time_start = time.time()
            timeout = 120
            while time.time() - time_start < timeout:
                time.sleep(1)
                if second_factor is not None:
                    break
            return second_factor

        def show_second_factor_dialog(msg: str):
            nonlocal second_factor
            second_factor = None
            if msg == Msg.code:
                output = run_dialog(self, "Verification code", password=False)
                second_factor = "" if output is None else output

        specific_protocol_args = {
            protocols.sftp.Protocol: {"two_factor_callback": second_factor_callback}
        }
        protocol_type = protocols.protocols_by_name[
            self.app_data.transfer_protocol_name
        ]
        protocol = protocol_type(
            **self.app_data.transfer_protocol_args[
                self.app_data.transfer_protocol_name
            ],
            **specific_protocol_args.get(protocol_type, {}),
        )

        worker = Worker(
            transfer.transfer,
            files=self.app_data.transfer_files,
            capture_loggers=(transfer.logger, protocols.sftp.logger),
            protocol=protocol,
            config=self.app_data.config,
            dry_run=dry_run,
            verify_dtr=self.app_data.transfer_verify_dtr,
            progress=progress,
            ignore_exceptions=True,
            report_config=self.app_data.config,
        )
        self.add_worker_actions(worker)
        msg_signal.msg.connect(show_second_factor_dialog)
        self.threadpool.start(worker)


class ConnectionDialog(QtWidgets.QDialog):
    def __init__(
        self,
        parent: Optional[QtWidgets.QWidget],
        label: str,
        forbidden: Sequence[str] = (),
        default_input: str = "",
    ):
        super().__init__(parent=parent)
        self.setWindowTitle("Connection profile")
        self.setWindowFlags(
            self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint  # type: ignore
        )

        btn_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        )
        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)

        self.text_field = LineEdit()
        self.text_field.setValidator(
            QtGui.QRegularExpressionValidator(QtCore.QRegularExpression(r"[\w\-@]+"))
        )
        self.text_field.setText(default_input)

        set_ok_enabled = lambda text: btn_box.button(
            QtWidgets.QDialogButtonBox.Ok
        ).setEnabled(len(text) > 0 and text not in forbidden)

        set_ok_enabled(self.text_field.text())
        self.text_field.textChanged.connect(set_ok_enabled)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(QtWidgets.QLabel(label))
        layout.addWidget(self.text_field)
        layout.addWidget(btn_box)
        self.setLayout(layout)


def get_btn_from_group(btn_grp: QtWidgets.QButtonGroup, text: str):
    btn = next(x for x in btn_grp.buttons() if x.text() == text)
    if not btn:
        raise KeyError(f"No button maching '{text}' found in the group.")
    return btn



try:
    # pylint: disable=unused-import
    from PySide6 import QtCore, QtGui, QtWidgets
    from PySide6.QtGui import QAction
except ImportError:
    from PySide2 import QtCore, QtGui, QtWidgets  # type: ignore
    from PySide2.QtWidgets import QAction  # type: ignore



import os

from ..core.error import UserError
from ..utils.config import create_config as _create_config, get_config_file
from ..utils.log import create_logger

logger = create_logger(__name__)


def create():
    """Creates a new default config file in the users config dir"""
    config_file = get_config_file()
    if os.path.isfile(config_file):
        raise UserError(f"The config file already exists at '{config_file}'")
    _create_config()
    logger.info("Created config file at '%s'", config_file)



import inspect
from typing import Callable, Type, Dict, Tuple

from .protocol import Protocol
from . import liquid_files, sftp
from ..utils.progress import ProgressInterface

protocols: Tuple[Type[Protocol], ...] = (sftp.Protocol, liquid_files.Protocol)
protocols_by_name: Dict[str, Type[Protocol]] = {
    p.__module__.replace(__name__ + ".", ""): p for p in protocols
}
__all__ = tuple(protocols_by_name)


def parse_protocol(s: str) -> Type[Protocol]:
    try:
        return protocols_by_name[s]
    except KeyError:
        raise ValueError(f"Invalid protocol: {s}") from None



from datetime import datetime
from functools import partial

from .component import (
    TabMixin,
    PathInput,
    GuiProgress,
    create_slider,
    run_dialog,
    SelectionButton,
    MandatoryLabel,
    create_verify_dtr_checkbox,
    LineEdit,
)
from .file_selection_widget import DirectoryAndFileSelectionWidget
from .model import TableModel
from .parallel import Worker
from .pyside import QtCore, QtGui, QtWidgets
from .settings_tab import LABEL_COMPRESSION_LEVEL, STATUS_TIP_COMPRESSION_LEVEL
from ..core.metadata import Purpose
from ..workflows import encrypt


class EncryptTab(QtWidgets.QWidget, TabMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.threadpool = QtCore.QThreadPool.globalInstance()
        self.app_data = self.parent().app_data

        files_panel = self.create_files_panel()
        self.create_run_panel(
            "Package and encrypt data", self.encrypt, "Package && Encrypt"
        )
        self.app_data.add_listener("encrypt_files", self._enable_buttons)
        self.app_data.add_listener("encrypt_recipients", self._enable_buttons)
        self.create_console()
        self.create_progress_bar()

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(files_panel)
        layout.addLayout(self._create_middle_layout())
        layout.addWidget(self.run_panel)
        layout.addWidget(self.console)
        layout.addWidget(self.progress_bar)

    def _enable_buttons(self):
        self.set_buttons_enabled(
            len(self.app_data.encrypt_recipients) > 0
            and len(self.app_data.encrypt_files) > 0
        )

    def create_files_panel(self):
        box = DirectoryAndFileSelectionWidget(
            "Files and/or directories to encrypt (hint: you can drag & drop files)",
            self,
        )
        box.file_list_model.layoutChanged.connect(
            lambda: setattr(self.app_data, "encrypt_files", box.get_list())
        )
        return box

    def create_user_panel(self):
        group_box = QtWidgets.QGroupBox("Keys")
        sender_widget = QtWidgets.QComboBox()
        sender_widget.setModel(self.app_data.priv_keys_model)
        if self.app_data.default_key_index:
            sender_widget.setCurrentIndex(self.app_data.default_key_index)

        recipients_input_view = QtWidgets.QComboBox()
        recipients_input_view.setModel(self.app_data.pub_keys_model)

        recipients_output_view = QtWidgets.QTableView()
        recipients_output_model = TableModel(columns=("Name", "Email", "Fingerprint"))
        recipients_output_view.setModel(recipients_output_model)
        recipients_output_view.verticalHeader().hide()
        recipients_output_view.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeToContents
        )
        recipients_output_view.horizontalHeader().setStretchLastSection(True)
        recipients_output_view.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        recipients_output_view.setSelectionMode(
            QtWidgets.QTableView.SelectionMode.SingleSelection
        )

        recipients_btn_add = QtWidgets.QPushButton(
            QtGui.QIcon(":icon/feather/user-plus.png"), ""
        )
        recipients_btn_add.setToolTip("Add recipient to the list")
        recipients_btn_remove = SelectionButton(
            "", recipients_output_view.selectionModel()
        )
        recipients_btn_remove.setIcon(QtGui.QIcon(":icon/feather/user-minus.png"))
        recipients_btn_remove.setToolTip("Remove recipient from the list")

        def update_sender(index):
            self.app_data.encrypt_sender = (
                ""
                if index == -1
                else self.app_data.priv_keys_model.get_value(index).fingerprint
            )

        def update_recipients(index):
            self.app_data.encrypt_recipients = [x[2] for x in index.model().get_data()]

        def add_recipient():
            key = recipients_input_view.model().get_value(
                recipients_input_view.currentIndex()
            )
            row = (key.uids[0].full_name, key.uids[0].email, key.fingerprint)
            recipients_output_model.set_data(
                set(recipients_output_model.get_data()) | set([row])
            )

        def remove_recipient():
            recipients_output_view.model().removeRows(
                recipients_output_view.currentIndex().row(), 1
            )
            recipients_output_view.clearSelection()

        # Connect actions
        recipients_btn_add.clicked.connect(add_recipient)
        recipients_btn_remove.clicked.connect(remove_recipient)
        recipients_output_model.dataChanged.connect(update_recipients)
        sender_widget.currentIndexChanged.connect(update_sender)
        # Set the default value for the sender
        update_sender(sender_widget.currentIndex())

        layout = QtWidgets.QGridLayout(group_box)
        layout.addWidget(
            QtWidgets.QLabel("Select data sender and add at least one recipient"),
            0,
            0,
            1,
            3,
        )
        layout.addWidget(MandatoryLabel("Sender"), 1, 0, 1, 1)
        layout.addWidget(sender_widget, 1, 1, 1, 1)
        layout.addWidget(recipients_input_view, 2, 1, 1, 1)
        layout.addWidget(recipients_btn_add, 2, 2, 1, 1)
        layout.addWidget(MandatoryLabel("Recipients"), 3, 0, 1, 1, QtCore.Qt.AlignTop)
        layout.addWidget(recipients_output_view, 3, 1, 1, 1)
        layout.addWidget(
            recipients_btn_remove, 3, 2, 1, 1, alignment=QtCore.Qt.AlignTop
        )

        return group_box

    def _create_middle_layout(self):
        layout_right = QtWidgets.QVBoxLayout()
        layout_right.addWidget(self._create_metadata_group_box())
        layout_right.addWidget(self._create_output_group_box())
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.create_user_panel())
        layout.addLayout(layout_right)
        return layout

    def _create_output_group_box(self):
        group_box = QtWidgets.QGroupBox("Output")
        # Create fields
        output_suffix = LineEdit()
        output_suffix.setStatusTip(
            "(optional) File name suffix in the " "format datetime_suffix.tar"
        )
        output_suffix.setValidator(
            QtGui.QRegularExpressionValidator(QtCore.QRegularExpression(r"[\w-]*"))
        )
        output_location = PathInput(path=self.app_data.encrypt_output_location)
        output_location.status_tip = "Destination folder for the encrypted package"
        # Add actions
        output_suffix.editingFinished.connect(
            lambda: setattr(
                self.app_data, "encrypt_output_suffix", output_suffix.text()
            )
        )
        output_location.on_path_change(
            partial(setattr, self.app_data, "encrypt_output_location")
        )

        slider_widgets = create_slider(
            minimum=0,
            maximum=9,
            initial_value=self.app_data.encrypt_compression_level,
            status_tip=STATUS_TIP_COMPRESSION_LEVEL,
            on_change=lambda n: setattr(self.app_data, "encrypt_compression_level", n),
        )
        compression_layout = QtWidgets.QHBoxLayout()
        for w in (QtWidgets.QLabel(LABEL_COMPRESSION_LEVEL),) + slider_widgets:
            compression_layout.addWidget(w)

        layout = QtWidgets.QGridLayout(group_box)
        layout.addWidget(QtWidgets.QLabel("Suffix"), 0, 0, 1, 1)
        layout.addWidget(output_suffix, 0, 1, 1, 1)
        layout.addWidget(QtWidgets.QLabel("Location"), 1, 0, 1, 1)
        layout.addWidget(output_location.text, 1, 1, 1, 1)
        layout.addWidget(output_location.btn, 1, 2, 1, 1)
        layout.addLayout(compression_layout, 2, 0, 1, 2)
        return group_box

    def _create_metadata_group_box(self):
        group_box = QtWidgets.QGroupBox("Metadata")
        # Create fields
        transfer_id = LineEdit()
        transfer_id.setStatusTip("(optional) Data Transfer Request ID")
        transfer_id.setValidator(QtGui.QIntValidator(1, 10 ** 9))
        verify_dtr = create_verify_dtr_checkbox(
            self.app_data, "encrypt_verify_dtr", group_box
        )
        purpose = QtWidgets.QComboBox(group_box)
        purpose.setStatusTip("(optional) Purpose of the package")
        purpose.addItems(("",) + tuple(x.name for x in Purpose))
        purpose.setCurrentText("")

        # Add actions
        def on_transfer_id_changed(text):
            self.app_data.encrypt_transfer_id = int(text) if text else None

        transfer_id.textChanged.connect(on_transfer_id_changed)
        purpose.currentTextChanged.connect(
            lambda text: setattr(
                self.app_data, "encrypt_purpose", None if text == "" else Purpose(text)
            )
        )
        layout = QtWidgets.QGridLayout(group_box)
        layout.addWidget(QtWidgets.QLabel("DTR ID"), 0, 0, 1, 1)
        layout.addWidget(transfer_id, 0, 1, 1, 1)
        layout.addWidget(verify_dtr, 1, 0, 1, 2)
        layout.addWidget(QtWidgets.QLabel("Purpose"), 2, 0, 1, 1)
        layout.addWidget(purpose, 2, 1, 1, 1)
        return group_box

    def encrypt(self, dry_run=False):
        progress = GuiProgress()
        progress.updated.connect(self.progress_bar.setValue)
        output_path = str(
            self.app_data.encrypt_output_location
            / "_".join(
                filter(
                    None,
                    [
                        datetime.now().astimezone().strftime(encrypt.DATE_FMT_FILENAME),
                        self.app_data.encrypt_output_suffix,
                    ],
                )
            )
        )
        if not dry_run and self.app_data.config.sign_encrypted_data:
            pw = run_dialog(self, "Enter password for your GPG key")
            if pw is None:
                return
        else:
            pw = None
        worker = Worker(
            encrypt.encrypt,
            self.app_data.encrypt_files,
            capture_loggers=(encrypt.logger,),
            sender=self.app_data.encrypt_sender,
            recipient=self.app_data.encrypt_recipients,
            dtr_id=self.app_data.encrypt_transfer_id,
            verify_dtr=self.app_data.encrypt_verify_dtr,
            config=self.app_data.config,
            passphrase=pw,
            output_name=output_path,
            dry_run=dry_run,
            offline=self.app_data.config.offline,
            compression_level=self.app_data.encrypt_compression_level,
            purpose=self.app_data.encrypt_purpose,
            progress=progress,
            ignore_exceptions=True,
            report_config=self.app_data.config,
        )
        self.add_worker_actions(worker)
        self.threadpool.start(worker)



import io
import os
import hashlib
from pathlib import Path
from typing import Tuple, Iterable, IO
from .error import UserError


def compute_checksum_sha256(file_object: IO[bytes], block_size: int = 65536) -> str:
    """Compute a sha256 checksum on a file by reading it by chunks of
    'block_size' bytes at a time.

    :param file_object: input file as a FileIO object.
    :param block_size: size of blocks (in bytes) to be returned by the
        generator function. The default block size is 2^16 = 65536 bytes.
    :return: checksum
    """
    hash_value = hashlib.sha256()
    block = file_object.read(block_size)
    while block:
        hash_value.update(block)
        block = file_object.read(block_size)
    return hash_value.hexdigest()


def compute_checksum_on_write(
    stream: IO[bytes],
    fout: io.FileIO,
    checksum_buffer: io.StringIO,
    block_size: int = 65536,
) -> None:
    """Compute a sha256 checksum on stream while writing it to a file at the
    same time.

    :param stream: a stream of bytes to checksum and write to disk.
    :param fout: output file object.
    :param checksum_buffer: a buffer to store the checksum.
    :param block_size: size of blocks (in bytes) to be returned by the
        generator function. The default block size is 2^16 = 65536 bytes.
    """
    hash_value = hashlib.sha256()
    block = stream.read(block_size)
    while block:
        hash_value.update(block)
        fout.write(block)
        block = stream.read(block_size)
    checksum_buffer.write(hash_value.hexdigest())
    checksum_buffer.seek(0)


def write_checksums(entries: Iterable[Tuple[str, IO[bytes]]]) -> bytes:
    """Returns bytes string containing lines in the format
    `<checksum> <file name>`
    (can directly be used by shasum: sha256sum --check *.sha256)

    Note that **DOS** path separators will be replaced by **Unix** ones.

    :param entries: Tuples of the form (archive_path, file like object)
    :return: The bytes string containing file names and hashes
    Here is an example:

    a7186ae7ff993b379qcf3567775cfc71a212rf217e4dd testDir/file1.fastq
    f8d2d394264823e711fgc34e4ac83f8cbc253c6we034f testDir/file2.fastq
    78f3b23fe49cf5f7f245ddf43v9788d9e62c0971fe5fb testDir/subdir2/file4.fastq
    :raises UserError: if backslash is used for archive name (on POSIX systems)
    """
    lines = []
    for archive_path, opener in entries:
        with opener as f_hash:
            checksum = compute_checksum_sha256(file_object=f_hash)
        if os.path.sep == "/" and "\\" in archive_path:
            raise UserError("On POSIX systems, backslashes are NOT allowed.")
        # Replace DOS path separators with Unix ones
        posix_path = Path(archive_path).as_posix()
        lines.append((f"{checksum} {posix_path}\n").encode())
    return b"".join(lines)


def verify_checksums(checksums: Iterable[Tuple[str, str]], base_path: str) -> None:
    """Checks that the checksum values of files listed in the 'checksums'
    generator match their hash values (also provided in 'checksums'). The
    function raises an error if a mismatch is detected.

    :param checksums: tuples of type (checksum value, file path).
    :param base_path: path of the directory containing the files to check.
    :raises UserError:
    """
    for checksum, file_name in checksums:
        # Open file in read-only mode and verify its checksum is the same
        # as the one given in input_file. File name should be posix, just in case it is NOT...
        file_path = Path(base_path) / Path(file_name.replace("\\", os.path.sep))
        with open(file_path, "rb") as f:
            computed = compute_checksum_sha256(f)
            checksum = checksum.lower()
            if checksum != computed:
                raise UserError(
                    f"Checksum mismatch for: {file_name}\nExpected: {checksum}"
                    f"\nGot: {computed}"
                )


def read_checksum_file(f: IO[bytes]) -> Iterable[Tuple[str, str]]:
    """Reads lines from a file object and parses them in the form:
     a7186ae7ff993b379qcf3567775cfc71a212rf217e4dd testDir/file1.fastq
     f8d2d394264823e711fgc34e4ac83f8cbc253c6we034f testDir/file2.fastq
     78f3b23fe49cf5f7f245ddf43v9788d9e62c0971fe5fb testDir/subdir2/file4.fastq

    The reason the input argument is a file object and not a file path is so
    that the function is more flexible and also allows to read input from
    a stream rather than an actual file.

    :param f: file object.
    :return: generator of tuples of the form (checksum value, file path).
    :raises UserError:
    """
    for line in f:
        # Split line into checksum and file name values. Each line must have
        # exactly 2 elements.
        try:
            checksum, file_name = line.decode().rstrip("\n\r").split(maxsplit=1)
        except ValueError:
            raise UserError("Input must have exactly 2 elements per line.") from None
        if os.path.isabs(file_name):
            raise UserError("Absolute path in checksum file")
        yield checksum, file_name



#!/usr/bin/env python3
# The {} placeholders in the docstring are to be replaced with URL_HELP
# and URL_GITLAB_ISSUES whenever the docstring is used.
"""Secure Encryption and Transfer Tool
For detailed documentation see: {}
To report an issue, please use: {}
"""

import os
import json
import shlex
import subprocess  # nosec B404:blacklist import_subprocess
from functools import wraps
from getpass import getpass
from typing import List, Dict, Any, Optional, Callable

from .progress import CliProgress
from .. import VERSION_WITH_DEPS
from ..utils.config import Config, load_config, config_to_dict
from .cli_builder import (
    rename,
    return_to_stdout,
    partial,
    lazy_partial,
    Subcommands,
    Subcommand,
    SubcommandGroup,
    decorate,
)
from ..utils.log import (
    log_to_stream,
    log_to_rotating_file,
    create_logger,
    log_runtime_info,
)
from ..utils.error_handling import (
    exit_on_exception,
    log_exceptions,
    error_report_hint_at_exit,
)
from ..workflows.config import create as create_config
from ..workflows.transfer import transfer as workflows_transfer
from ..workflows.decrypt import decrypt as workflows_decrypt
from ..workflows.encrypt import encrypt as workflows_encrypt
from ..workflows.request_sigs import request_sigs
from ..workflows.upload_keys import upload_keys
from ..protocols import Protocol, parse_protocol, __all__ as available_protocols, sftp
from ..core.versioncheck import check_version
from ..core.error import UserError
from .. import URL_READTHEDOCS, URL_GITLAB_ISSUES

logger = create_logger(__name__)


def parse_dict(s):
    return json.loads(s)


def parse_protocol_args(s):
    args = parse_dict(s)
    not_provided = object()
    pw = args.get("pkey_password", not_provided)
    if pw is None:  # pw is provided and is None
        args["pkey_password"] = getpass("Please enter your ssh private key password:")
    return args


def two_factor_cli_prompt():
    return input("Verification code: ")


def get_passphrase_from_cmd(passphrase_cmd: str) -> str:
    try:
        # B603:subprocess_without_shell_equals_true
        # Suppress the warning, user can easily execute the same command
        # directly in the shell.
        proc = subprocess.run(  # nosec
            shlex.split(passphrase_cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
        return proc.stdout.decode().strip()
    except subprocess.CalledProcessError as e:
        raise UserError(
            "Failed to read passphrase from "
            f"'{passphrase_cmd}'. "
            f"{e.stderr.decode().strip()}"
        ) from e


def get_passphrase_from_cmd_or_prompt(msg: str, passphrase_cmd: Optional[str]) -> str:
    if passphrase_cmd:
        return get_passphrase_from_cmd(passphrase_cmd)
    return getpass(msg)


@wraps(workflows_encrypt)
def encrypt(
    *args,
    config: Config,
    dry_run: bool,
    passphrase_cmd: Optional[str],
    dtr_id: Optional[int] = None,
    offline: bool = False,
    verify_dtr: Optional[bool] = None,
    **kwargs,
):
    if verify_dtr is None:
        verify_dtr = not offline and dtr_id is not None
    if offline and verify_dtr:
        raise UserError(
            "--verify_dtr is not possible when offline is " "enabled in the config"
        )
    if config.sign_encrypted_data and not dry_run:
        kwargs["passphrase"] = get_passphrase_from_cmd_or_prompt(
            "Please enter your gpg private key password:", passphrase_cmd
        )
    else:
        kwargs["passphrase"] = None
    return workflows_encrypt(
        *args,
        dtr_id=dtr_id,
        offline=offline,
        verify_dtr=verify_dtr,
        config=config,
        dry_run=dry_run,
        **kwargs,
    )


@wraps(workflows_decrypt)
def decrypt(*args, dry_run: bool, passphrase_cmd: Optional[str], **kwargs):
    if dry_run:
        kwargs["passphrase"] = None
    else:
        kwargs["passphrase"] = get_passphrase_from_cmd_or_prompt(
            "Please enter your gpg private key password:", passphrase_cmd
        )
    return workflows_decrypt(*args, dry_run=dry_run, **kwargs)


def transfer(
    files: List[str],
    *,
    connection: str = None,
    two_factor_callback,
    passphrase_cmd: Optional[str] = None,
    config: Config,
    protocol=None,
    protocol_args: Dict[str, Any] = None,
    dry_run: bool = False,
    verify_dtr: Optional[bool] = None,
    progress=None,
):
    if connection is not None:
        if protocol is not None:
            raise UserError(
                "Arguments 'protocol' and 'connection' " "cannot be given together"
            )
        connection_obj = config.connections[connection]
        protocol = parse_protocol(connection_obj.protocol)
        if protocol_args is None:
            protocol_args = {}
        protocol_args = {**connection_obj.parameters, **protocol_args}
    if protocol is None or protocol_args is None:
        raise UserError(
            "Either 'protocol' together with 'protocol_args' "
            "or 'connection' "
            "has to be given as an argument"
        )
    specific_protocol_args = {
        sftp.Protocol: {"two_factor_callback": two_factor_callback}
    }
    try:
        protocol_instance: Protocol = protocol(
            **protocol_args, **specific_protocol_args.get(protocol, {})
        )
    except TypeError as e:
        raise UserError(format(e).replace("__init__", protocol.__module__)) from None
    for pw_arg in protocol_instance.required_password_args():
        if protocol_args.get(pw_arg) is None:
            setattr(
                protocol_instance,
                pw_arg,
                get_passphrase_from_cmd_or_prompt(
                    f"Please enter the password for the argument `{pw_arg}`:",
                    passphrase_cmd,
                ),
            )
    return workflows_transfer(
        files,
        protocol=protocol_instance,
        config=config,
        dry_run=dry_run,
        verify_dtr=verify_dtr,
        progress=progress,
    )


def load_config_check():
    cfg = load_config()
    if not cfg.offline and cfg.check_version:
        check_version(cfg.repo_url)
    return cfg


class Cli(Subcommands):
    description = __doc__.format(URL_READTHEDOCS, URL_GITLAB_ISSUES)
    version = VERSION_WITH_DEPS
    config = load_config_check()
    passphrase_override = dict(
        help="Instead of asking for passphrase, "
        "get it from an external command "
        "(passphrase must be returned to the "
        "standard output).",
        name="passphrase-cmd",
        dest="passphrase_cmd",
    )
    dry_run_override = dict(
        help="Perform checks on the input data, without running the actual command."
    )
    verify_dtr_override = dict(
        help="Verify DTR ID against online database.", default=None
    )
    subcommands = (
        Subcommand(
            decorate(
                encrypt,
                partial(config=config, offline=config.offline),
                lazy_partial(progress=CliProgress),
            ),
            overrides={
                "files": dict(help="Input file(s) or directories."),
                "sender": dict(
                    help="fingerprint, key ID or email associated "
                    "with GPG key of data sender.",
                    alias="-s",
                ),
                "recipient": dict(
                    help="fingerprint, key ID or email associated with "
                    "GPG key of data recipient(s).",
                    alias="-r",
                ),
                "dtr_id": dict(
                    help="Data Transfer Request (DTR) ID (optional).", alias="-t"
                ),
                "verify_dtr": verify_dtr_override,
                "purpose": dict(
                    help="Purpose of the DTR (PRODUCTION, TEST)."
                    "Mandatory only if `transfer_id` is specified."
                ),
                "output_name": dict(
                    help="output encrypted file name. If no path is specified, "
                    "the output tarball is saved in the current working "
                    "directory. If this argument is "
                    "missing the output name is set to a string based on "
                    "the current date and time the function.",
                    default=None,
                    alias="-o",
                ),
                "dry_run": dry_run_override,
                "compression_level": dict(
                    help="compression level for inner tarball in the range "
                    "0 (no compression) to 9 (highest compression). "
                    "Higher compression levels require more computing "
                    "time."
                ),
                "passphrase": passphrase_override,
            },
        ),
        Subcommand(
            decorate(
                transfer,
                partial(config=config, two_factor_callback=two_factor_cli_prompt),
                lazy_partial(progress=CliProgress),
            ),
            overrides={
                "files": dict(help="Input file(s) or directories"),
                "protocol": dict(
                    help="The protocol for the file transfer."
                    "Currently available: {}".format(", ".join(available_protocols)),
                    type=parse_protocol,
                    alias="-p",
                ),
                "protocol_args": dict(
                    help="Protocol specific arguments. "
                    "Must be passed as a json string",
                    type=parse_protocol_args,
                ),
                "connection": dict(
                    help="Instead of the option 'protocol', load a connection "
                    "named by the argument of this option to use the "
                    "protocol and protocol args from the config. The "
                    "protocol args can be overwritten by the "
                    "protocol_args option"
                ),
                "passphrase_cmd": dict(help=passphrase_override["help"]),
                "dry_run": dry_run_override,
                "verify_dtr": verify_dtr_override,
            },
        ),
        Subcommand(
            decorate(
                decrypt, partial(config=config), lazy_partial(progress=CliProgress)
            ),
            overrides={
                "output_dir": dict(
                    help="Output directory.", default=os.getcwd(), alias="-o"
                ),
                "decrypt_only": dict(help="Skip extraction."),
                "dry_run": dry_run_override,
                "passphrase": passphrase_override,
            },
        ),
        SubcommandGroup(
            "config",
            decorate(
                config_to_dict, rename("show"), return_to_stdout, partial(config=config)
            ),
            create_config,
            help="Commands related to config file",
        ),
        Subcommand(
            decorate(request_sigs, partial(config=config)),
            overrides={"key_ids": dict(help="Key ids to request signatures for")},
        ),
        Subcommand(
            decorate(upload_keys, partial(config=config)),
            overrides={"key_ids": dict(help="Key ids to upload")},
        ),
    )


def when(
    condition: bool, decorator: Callable[[Callable], Callable]
) -> Callable[[Callable], Callable]:
    if condition:
        return decorator
    return lambda x: x


log_to_rotating_file(
    log_dir=Cli.config.log_dir, file_max_number=Cli.config.log_max_file_number
)
log_to_stream()


@exit_on_exception
@when(Cli.config.error_reports, error_report_hint_at_exit(Cli.config))
@log_exceptions(logger)
def run():
    log_runtime_info(logger)
    if Cli():
        return 0
    return 1


if __name__ == "__main__":
    run()



from typing import List, Tuple, Dict, Optional
import json

from ..utils.config import Config
from ..utils.log import create_logger, log_timing
from ..utils.progress import ProgressInterface
from ..core.crypt import (
    get_recipient_email,
    extract_pub_key_ids,
    validated_keys_by_ids,
    gpg,
)
from ..core.archive import check_tar, extract, METADATA_FILE, DATA_FILE_ENCRYPTED
from ..core.metadata import load_metadata
from ..core.error import UserError
from ..protocols import Protocol


logger = create_logger(__name__)


@log_timing(logger)
def transfer(
    files: List[str],
    *,
    protocol: Protocol,
    config: Config,
    dry_run: bool = False,
    verify_dtr: Optional[bool] = None,
    progress: ProgressInterface = None,
):
    """Transfer file(s) to the selected recipient on the BiomedIT network."""

    logger.info(
        """Input summary:
    files to transfer: %s"
    dry run          : %s""",
        "\n\t".join(files),
        str(dry_run),
    )

    with logger.log_task("Input files verification"):
        for tar in files:
            check_tar(tar)

    files_by_recipient: Dict[Tuple[gpg.Key, ...], List[str]] = {}
    with logger.log_task("Extracting destination for each file to transfer"):
        for tar in files:
            logger.info("\t-> %s", tar)
            with extract(open(tar, "rb"), METADATA_FILE, DATA_FILE_ENCRYPTED) as (
                metadata,
                encrypted_file,
            ):
                metadata = json.load(metadata)
                keys = tuple(
                    validated_keys_by_ids(
                        extract_pub_key_ids(encrypted_file),
                        config.gpg_store,
                        config.key_validation_authority_keyid,
                        config.keyserver_url,
                    )
                )
            metadata = load_metadata(metadata)
            if verify_dtr is None:
                verify_dtr = metadata.transfer_id is not None
            if verify_dtr:
                if metadata.transfer_id is None:
                    raise UserError(
                        "DTR (Data Transfer Request) ID is " "missing in file metadata."
                    )
                try:
                    project_id = config.portal_api.verify_transfer(
                        metadata=metadata, filename=tar
                    )
                    logger.info(
                        "DTR ID '%s' is valid for project '%s'",
                        metadata.transfer_id,
                        project_id,
                    )
                except RuntimeError as e:
                    raise UserError(format(e)) from e
            files_by_recipient.setdefault(keys, []).append(tar)

    if dry_run:
        logger.info("Dry run completed successfully")
        return

    for recipient_keys, r_files in files_by_recipient.items():
        emails = [get_recipient_email(k) for k in recipient_keys]
        if hasattr(protocol, "recipients"):
            setattr(protocol, "recipients", emails)
        if hasattr(protocol, "pkey_password_encoding"):
            setattr(protocol, "pkey_password_encoding", config.ssh_password_encoding)
        with logger.log_task(
            "Transferring files encrypted for " f"{', '.join(emails)}"
        ):
            protocol.upload(r_files, progress=progress)



import mimetypes
import uuid
from typing import Tuple


def encode(name: str, filename: str, data: bytes, extra_data=()) -> Tuple[str, bytes]:
    boundary = uuid.uuid4().hex
    content_type = "multipart/form-data; boundary={}".format(boundary)
    boundary_b = boundary.encode()
    inner_content_type = (
        mimetypes.guess_type(filename)[0] or "application/octet-stream"
    ).encode()
    name_b = name.encode()
    filename_b = filename.encode()
    boundary_token = b"--" + boundary_b + b"\r\n"
    return (
        content_type,
        b"".join(
            map(lambda key_val: boundary_token + encode_value(*key_val), extra_data)
        )
        + b"--"
        + boundary_b
        + b"\r\n"
        b'Content-Disposition: form-data; name="'
        + name_b
        + b'"; filename="'
        + filename_b
        + b'"\r\n'
        b"Content-Type: " + inner_content_type + b"\r\n"
        b"\r\n" + data + b"\r\n--" + boundary_b + b"--\r\n",
    )


def encode_value(key: str, value) -> bytes:
    return (
        b'Content-Disposition: form-data; name="' + key.encode() + b'"\r\n'
        b"\r\n" + str(value).encode() + b"\r\n"
    )



import io
import os
from contextlib import contextmanager
from typing import List, Iterable, Tuple, Callable, Optional, Sequence, IO, cast
from collections.abc import Collection
from itertools import repeat
from functools import wraps


class ProgressInterface:
    """Interface for progress objects."""

    def update(self, completed_fraction: float) -> None:
        """Method to update the status of a progress object to a new
        fraction of completeness.
        :param float completed_fraction: fraction of completion value
            must be 0 <= completed_fraction <= 1.
        """

    def get_completed_fraction(self) -> float:
        """Property to get the current completed fraction of the progress
        interface, i.e. the current fraction of completion of the task.
        """


def progress_iter(
    iterable: Iterable,
    progress: ProgressInterface = None,
    weights: List[float] = None,
    iterable_length: int = None,
) -> Iterable:
    """Add progress tracking through a ProgressInterface to an iterable object.
    :param iterable: iterable object to which progress tracking is added.
    :param progress: progress interface objects that will take care of
        progress tracking.
    :param weights: weights to be associated with each iteration of the input
        iterable object. If weights are not specified, all iterations are
        assumed to have equal weight.
    :param iterable_length: length of the iterable object. Only needed if the
        iterable does not have a __len__() method.
    :return: Yields from :iterable:
    """
    # Case 1: no progress interface was passed => no progress to track.
    if progress is None:
        yield from iterable

    # Case 2: Iterate over the input iterable wile updating the progress
    # interface. Essentially this loops (yield) through the iterable, and
    # at each iteration calls the update method on the ProgressInterface
    # object.
    # At the end, send signal that completion reached 100% to the progress
    # interface.
    else:
        yield from _progress_iter_common(
            _progress_wrapper, iterable, progress, weights, iterable_length
        )
        progress.update(1.0)


def progress_iter_with_sub_progress(
    iterable: Iterable,
    progress: Optional[ProgressInterface] = None,
    weights: List[float] = None,
    iterable_length: int = None,
) -> Iterable:
    """Add progress tracking through a ProgressInterface to an iterable object.
    The difference with 'progress_iter' is that items of the returned iterable
    will now contain tuples consisting of the original item and a
    ProgressInterface object which tracks progress of the subtasks within the
    whole progress range.
    For input parameters, see 'progress_iter' function.
    """
    # If no progress interface was passed => no progress to track and the
    # function simply returns an iterable consisting of tuples of the input
    # iterable values and 'None'.
    if progress is None:
        return zip(iterable, repeat(None))

    # Return tuples of the original iterable values and an associated
    # ProgressInterface.
    return _progress_iter_common(
        _progress_wrapper_with_sub_progress,
        iterable,
        progress,
        weights,
        iterable_length,
    )


def _progress_iter_common(
    wrapper: Callable,
    iterable: Iterable,
    progress: ProgressInterface = None,
    weights: Optional[Sequence[float]] = None,
    iterable_length: int = None,
) -> Iterable:
    """Iterate over all values in the input iterable and apply the specified
    wrapper function to them. The wrapper function must accept arguments that
    have the form (item, (start_position, stop)).
    """
    # Get length of input iterable.
    if isinstance(iterable, Collection):
        iterable_length = iterable.__len__()
    elif iterable_length is None:
        raise ValueError(
            "iterable has no length attribute and no value was"
            "passed for parameter [iterable_length] either."
        )

    # If no weight were provided, assume equal weights for all iterations.
    if weights is None:
        weights = (1,) * iterable_length

    # Create sub-ranges of the 0-1 range proportionally to the input weights.
    ranges = weights_to_ranges(weights)

    # Iterate over input iterator while each time calling the update method
    # of the associated progress interface. map applies the update_progress
    # function to all elements returned by zip.
    update_progress = wrapper(progress)
    return map(update_progress, zip(iterable, ranges))


def weights_to_ranges(
    weights: Sequence[float], start: float = 0.0
) -> Iterable[Tuple[float, float]]:
    """Splits the range (0.0, 1.0) into a number of sub-ranges equal to the
    number of weights, with the size of each sub-range proportional to the
    value of its weights.
    For instance, if weights = [2,1,1], the range 0-1 is split into 3 sub-
    ranges, where the first has twice the size of the second and third:
    (0.0, 0.5), (0.5, 0.75), (0.75, 1.0).
    """
    # Ensure that empty files are > 0 for progress computation
    safe_weights = [w or 1.0 for w in weights]
    sum_of_weights = sum(safe_weights)
    stop = safe_weights[0] / sum_of_weights
    yield (start, stop)
    for weight in safe_weights[1:]:
        start = stop
        stop += weight / sum_of_weights
        yield (start, stop)


def _progress_wrapper(progress: ProgressInterface) -> Callable:
    """Create a wrapper function that takes an 'item_with_range' input object
    and returns the bare input item (an iterable) and updates the progress
    interface with the 'start' position of the range associated to the item.
    An 'item_with_range' is a tuple that has the form (iterable, range), where
    iterable is the iterable for which we track progress and range is a tuple
    (start, stop) giving the range of completion fractions for the iterable.
    """

    def progress_wrapper(
        item_with_range: Tuple[Iterable, Tuple[float, float]]
    ) -> Iterable:
        """item_with_range: iterable and its completion range
        :return: The bare item
        """
        item, (start, _stop) = item_with_range
        progress.update(start)
        return item

    return progress_wrapper


def _progress_wrapper_with_sub_progress(progress: ProgressInterface) -> Callable:
    """Return a function that takes an 'item_with_range' and return the
    iterable item with an associated ProgressInterface that does automatic
    rescaling of its .update() method in the range of 'item_with_range'.
    """

    def progress_wrapper(
        item_with_range: Tuple[Iterable, Tuple[float, float]]
    ) -> Tuple[Iterable, Optional[ProgressInterface]]:
        """item_with_range: iterable and its completion range
        :return: A tuple consisting of the bare item and a ProgressInterface instance
        """
        iterable_item, (start, stop) = item_with_range
        sub_progress = scale_progress(progress, start, stop)
        return iterable_item, sub_progress

    return progress_wrapper


def scale_progress(
    progress: ProgressInterface, start: float, stop: float
) -> Optional[ProgressInterface]:
    """Create an instance of a ProgressInterface with a modified .update
    method that rescales the value of the 'completed_fraction' to be in the
    range (start, stop).
    Example: if we specify start=0.25, stop=0.5, the .update
    method of the new instance will rescale any 'completed_fraction' fraction
    value to the range 0.25 - 0.5. So calling .update(0.5) will in fact update
    the progress instance to 0.375 instead of 0.5, and update(1.0) will update
     the progress instance to 0.5 and not to 1.0.
    """
    if start > stop:
        raise ValueError("start > stop")
    if progress is None:
        return None

    # Define a new class that will be returned by the function. The .update
    # method is modified so it automatically rescales its inputs to the
    # desired range of values.
    class SubProgress(ProgressInterface):
        def update(self, completed_fraction: float) -> None:
            """Rescale the completed_fraction value so that its value is
            in the range [start-stop].
            """
            completed_fraction = min(completed_fraction, 1)
            progress.update(start + completed_fraction * (stop - start))

        def get_completed_fraction(self) -> float:
            """Rescale the completed fraction so that its value is in the
            range [start-stop].
            """
            completed_fraction = progress.get_completed_fraction()
            return (completed_fraction - start) / (stop - start)

    return SubProgress()


@contextmanager
def subprogress(progress: Optional[ProgressInterface], step_completion_increase: float):
    """Context manager for running a progress bar or a subset of a progress
    bar (i.e. a "scaled progress".
    Usage example:
    with subprogress(my_progress_bar, 0.3, 'step 2') as sub_progressbar:
        ...

    :param progress: progress interface for which to create a sub-progress.
    :param step_completion_increase: increase in completion associated to the
        current subprogress step of the progress interface. In example, if
        step_completion_increase = 0.3, then it means that the current
        subprogress step will contribute an increase of 0.3 to the completed
        fraction value of the progress interface. So if at the start of the
        subprogress step the completed fraction is 0.25, then at the end the
        value will be 0.55 (0.25 + 0.3).
    :return: A ProgressInterface instance covering the range starting at the
        current position of :progress: and advancing :step_completion_increase:
    :raises ValueError:
    """
    if progress is None:
        yield None
        return
    start = progress.get_completed_fraction()
    stop = start + step_completion_increase
    if stop > 1.0:
        raise ValueError(
            f"progress interface completion value cannot be > 1. "
            f"Value passed for 'completed_fraction_increase' "
            f"[{step_completion_increase}] is too large."
        )
    yield scale_progress(progress, start, stop)


class FileObjectProgress:
    """Wraps a file object

    All read operations will be tracked and passed to the ProgressInterface
    instance
    """

    def __init__(self, f_obj: IO[bytes], progress):
        self._f = f_obj
        original_pos = f_obj.tell()
        f_obj.seek(0, io.SEEK_END)
        total_size = f_obj.tell() - original_pos
        f_obj.seek(original_pos)

        if total_size == 0:

            def wrap_update(f):
                progress.update(1)
                return f

        else:

            def wrap_update(f):
                @wraps(f)
                def wrapped_f(*args, **kwargs):
                    data = f(*args, **kwargs)
                    progress.update((f_obj.tell() - original_pos) / total_size)
                    return data

                return wrapped_f

        for f_name in ("__next__", "read", "readline", "readlines"):
            f = getattr(f_obj, f_name)
            if f is not None:
                setattr(self, f_name, wrap_update(f))
        super().__init__()

    def __iter__(self):
        return self

    def __next__(self):
        # Note: self.__next__ != FileObjectProgress.__next__
        return self.__next__()

    def __enter__(self):
        self._f.__enter__()
        return self

    def __exit__(self, *args):
        return self._f.__exit__(*args)

    def __getattr__(self, attr):
        return getattr(self._f, attr)

    @property
    def name(self):
        return self.__getattr__("name")


def with_progress(f: IO[bytes], progress: Optional[ProgressInterface]) -> IO[bytes]:
    if progress is None:
        return f
    return cast(IO[bytes], FileObjectProgress(f, progress))


class Opener:
    """Wraps a file object

    Acts as a replacement for builtin `open` which only opens when
    `__enter__` is called on the object.
    This is needed when building a large list of file objects
    which in a first loop are only accessed for their file name
    and in a second loop are opened.
    Builtin `open` would already open the file on the open call,
    resulting in "OSError: [Errno 24] Too many open files" in
    the first loop
    """

    def __init__(self, f: str, mode: str = None, wrapper=lambda x: x):
        self.name = f
        self.mode = mode
        self._f = None
        self.wrapper = wrapper

    def __enter__(self):
        # Impossible to use `with` if you want to wrap `__enter__`, `__exit__`.
        # Workaround would be contextlib.contextmanager. But then you cannot set the
        # `name`, `mode` attributes on the contextmanager object without breaking mypy.
        self._f = open(self.name, self.mode)  # pylint: disable=consider-using-with
        return self.wrapper(self._f.__enter__())

    def __exit__(self, *args):
        return self._f.__exit__(*args)


def progress_file_iter(
    files: List[str], mode: str, progress: Optional[ProgressInterface]
) -> Iterable[IO[bytes]]:
    """Associate, to each input file, a sub-progress object whose contribution
    to the increase in completion of the input progress object is proportional
    to the file's size. The output is an iterator of file objects with their
    associated sub-progress already backed-in.
    Example: this can be used to track progress when reading through a list
    of files. Each file that is read will then contribute to the increment in
    progress proportionally to its size.

    :param files: list of files to which progress should be associated.
    :param mode: file access mode, e.g. 'rb' or 'r'.
    :param progress: ProgressInterface to which the file objects are associated.
    :return: file objects with backed-in subprogress.
    """
    file_size = [float(os.path.getsize(f)) for f in files]
    files_with_progress = progress_iter_with_sub_progress(
        iterable=files, progress=progress, weights=file_size, iterable_length=len(files)
    )
    for file, sub_progress in files_with_progress:
        yield cast(
            IO[bytes],
            Opener(file, mode=mode, wrapper=with_progress_decorator(sub_progress)),
        )


def with_progress_decorator(progress):
    def _wrapper(f):
        return with_progress(f=f, progress=progress)

    return _wrapper



import platform

try:
    from importlib.metadata import version as _version, PackageNotFoundError

    def version(dep: str):
        try:
            return _version(dep)
        except PackageNotFoundError:
            return None


except ImportError:
    from pkg_resources import get_distribution, DistributionNotFound

    def version(dep: str):
        try:
            return get_distribution(dep).version
        except DistributionNotFound:
            return None


APP_NAME_SHORT = "sett"
APP_NAME_LONG = "Secure Encryption and Transfer Tool"
URL_READTHEDOCS = "https://sett.rtfd.io"
URL_GITLAB = "https://gitlab.com/biomedit/sett"
URL_GITLAB_ISSUES = URL_GITLAB + "/-/issues"

__project_name__ = "sett"
__version__ = version(__project_name__) or "0.0.0.dev"


def dep_version(dep: str) -> str:
    return version(dep) or "not_found"


VERSION_WITH_DEPS = (
    f"{APP_NAME_SHORT} {__version__} ("
    + ", ".join(f"{n} {dep_version(n)}" for n in ("gpg-lite", "libbiomedit"))
    + ")"
)

RUNTIME_INFO = f"{VERSION_WITH_DEPS} (platform: {platform.platform()})"



import os
import json

from functools import partial
from typing import List, Optional

from ..core.error import UserError
from ..core.archive import (
    check_tar,
    extract,
    unpack_from_stream,
    METADATA_FILE,
    DATA_ARCHIVE,
    DATA_FILE_ENCRYPTED,
    CHECKSUM_FILE,
    extract_with_progress,
)
from ..core.crypt import (
    decrypt as core_decrypt,
    extract_pub_key_ids,
    validated_keys_by_ids,
    fingerprint2keyid,
    verify_metadata_signature,
)
from ..core.filesystem import DeleteDirOnError, unique_filename
from ..core.checksum import (
    verify_checksums,
    compute_checksum_sha256,
    read_checksum_file,
)
from ..core.metadata import MetaData
from ..core.secret import Secret
from ..utils.log import create_logger, log_timing
from ..utils.progress import ProgressInterface, subprogress
from ..utils.config import Config


logger = create_logger(__name__)


@log_timing(logger)
def decrypt(
    files: List[str],
    *,
    passphrase: Optional[Secret] = None,
    output_dir: str,
    config: Config,
    decrypt_only: bool = False,
    dry_run: bool = False,
    progress: Optional[ProgressInterface] = None,
):
    """Decrypt and decompress an input .tar file."""

    logger.info(
        """Input summary:
    file(s) to decrypt: %s
    output_dir: %s
    dry_run: %s
""",
        "\n\t".join(files),
        output_dir,
        dry_run,
    )

    with logger.log_task("Input data check"):
        for tar_name in files:
            check_tar(tar_name)

    if dry_run:
        logger.info("Dry run completed successfully")
        return

    for tar_file in files:
        # Reset progress for each tar file
        if progress is not None:
            progress.update(0.0)
        logger.info("Untar file %s", tar_file)
        decrypt_tar(
            tar_file,
            passphrase=passphrase,
            output_dir=output_dir,
            config=config,
            decrypt_only=decrypt_only,
            progress=progress,
        )
        logger.info("Data decryption completed successfully.")


def decrypt_tar(
    tar_file: str,
    passphrase: Optional[Secret],
    output_dir: str,
    config: Config,
    decrypt_only: bool = False,
    progress: Optional[ProgressInterface] = None,
):

    # To avoid overwriting files, each tarball is untarred in a directory
    # that has the same name as the tar file minus the .tar extension.
    out_dir = unique_filename(
        os.path.splitext(os.path.join(output_dir, os.path.basename(tar_file)))[0]
    )

    with DeleteDirOnError(out_dir):
        with logger.log_task("Verifying signatures..."):
            verify_metadata_signature(
                tar_file,
                config.gpg_store,
                config.key_validation_authority_keyid,
                config.keyserver_url,
            )

        with logger.log_task("Verifying encryption keys..."), open(
            tar_file, "rb"
        ) as tar_obj, extract(tar_obj, DATA_FILE_ENCRYPTED) as f_data:
            keys = validated_keys_by_ids(
                extract_pub_key_ids(f_data),
                config.gpg_store,
                config.key_validation_authority_keyid,
                config.keyserver_url,
            )
            logger.info(
                "Data encrypted for:\n%s",
                "\n".join(
                    f"User ID    : {key.uids[0]}\n" f"Fingerprint: {key.fingerprint}"
                    for key in keys
                ),
            )

        with logger.log_task("Verifying checksums..."), open(
            tar_file, "rb"
        ) as tar_obj, subprogress(
            progress, step_completion_increase=0.05
        ) as scaled_progress, extract_with_progress(
            tar_obj, scaled_progress, METADATA_FILE, DATA_FILE_ENCRYPTED
        ) as (
            f_metadata,
            f_data,
        ):
            metadata = MetaData.from_dict(json.load(f_metadata))
            if metadata.checksum.lower() != compute_checksum_sha256(f_data):
                raise UserError(f"Checksum mismatch for {f_data.name}")

        with logger.log_task("Decrypting data..."), open(
            tar_file, "rb"
        ) as tar_obj, subprogress(
            progress, step_completion_increase=0.9
        ) as scaled_progress, extract_with_progress(
            tar_obj, scaled_progress, DATA_FILE_ENCRYPTED
        ) as f_data:
            os.makedirs(out_dir, exist_ok=True)
            if decrypt_only:
                sender_fprs = core_decrypt(
                    source=f_data,
                    output=os.path.join(out_dir, DATA_ARCHIVE),
                    gpg_store=config.gpg_store,
                    passphrase=passphrase,
                )
            else:
                unpacked: List[str] = list()
                sender_fprs = core_decrypt(
                    source=f_data,
                    output=partial(unpack_from_stream, dest=out_dir, content=unpacked),
                    gpg_store=config.gpg_store,
                    passphrase=passphrase,
                )
                log_files(unpacked)
            sender_sig_keys = validated_keys_by_ids(
                map(fingerprint2keyid, sender_fprs),
                config.gpg_store,
                config.key_validation_authority_keyid,
                config.keyserver_url,
            )
            logger.info(
                "Data signed by:\n%s",
                "\n".join(
                    f"User ID    : {key.uids[0]}\n" f"Fingerprint: {key.fingerprint}"
                    for key in sender_sig_keys
                ),
            )

        with subprogress(progress, step_completion_increase=0.05) as scaled_progress:
            if not decrypt_only:
                with logger.log_task(
                    "Checksum verification of uncompressed data..."
                ), open(os.path.join(out_dir, CHECKSUM_FILE), "rb") as fout:
                    sha_checks = list(read_checksum_file(fout))
                    verify_checksums(sha_checks, out_dir)
            if scaled_progress is not None:
                scaled_progress.update(1.0)


def log_files(files: List[str]):
    max_files_to_list = 50
    d_n = len(files) - max_files_to_list
    logger.info(
        "List of extracted files: \n\t%s%s",
        "\n\t".join(files[:max_files_to_list]),
        ((f"\n\t and {d_n} more files " "- not listing them all.") if d_n > 0 else ""),
    )
